"""
This module contains data range definitions.
"""
from __future__ import annotations

import datetime as dt
import math
from abc import ABC, abstractmethod
from typing import Any, Dict, List

import numpy as np
import polars as pl
from pydantic import BaseModel, Field

__all__ = ["CyclicMinMax", "DataElement", "DataRange", "ListOfValues", "MinMax"]


class DataElement:
    """
    Wrapper class to create an instance of an object given by a dtype
    """

    @classmethod
    def create(cls, value, dtype):
        """
        Create a DataElement from a given value and d(ata)type

        :param value: value of the DataElement
        :param dtype: datatype of the value
        """
        if isinstance(dtype, pl.datatypes.classes.DataTypeClass):
            dtype = dtype.to_python()
        elif isinstance(dtype, pl.datatypes.Datetime):
            if type(value) == str:
                value = dt.datetime.fromisoformat(value)
            return dtype.to_python().fromtimestamp(value.timestamp())

        return dtype(value)


class DataRange(ABC):
    """
    Representation of a data range.

    This class should be subclassed to implement a custom data range
    """

    @abstractmethod
    def is_in_range(self, value) -> bool:
        """
        Check if a value lies in the permitted range.

        :param value: Value to check
        :return: True if value lies in range, False otherwise
        """

    def is_not_in_range(self, value) -> bool:
        return not self.is_in_range(value)

    def __contains__(self, value) -> bool:
        """
        Check if value lies in the given range via 'in':

        .. highlight:: python
        .. code-block:: python

            custom_range = MinMax(0.0, 10.0)
            if 1.0 in custom_range:
                ...

        :param value: Value to test
        :return: True if value is in range, False otherwise
        """

        return self.is_in_range(value=value)

    @abstractmethod
    def to_dict(self) -> str:
        """
        Convert object to dictionary to allow plain type serialisation.

        :raise NotImplementedError: If method has not been implemented by a subclass
        """
        return dict(self)

    def __iter__(self):
        """
        Convert object to dictionary to allow plain type serialisation.

        :raise NotImplementedError: If method has not been implemented by a subclass
        """
        raise NotImplementedError(f"{self.__class__.__name__}.__iter__ not implemented")

    @classmethod
    def from_dict(cls, data: Dict[str, Any], dtype: Any = None) -> DataRange:
        """
        Load the data range from a plain type description in a dictionary.

        .. highlight:: python
        .. code-block:: python

            min_max_range = DataRange.from_dict({ "MinMax": { "min": 0, "max": 1 }}, dtype=float)

        :param data: dictionary data
        :param dtype: the value type to use for initialisation,
        :return: The newly created class
        :raise ValueError: Raises if the requested range type is not known
        """
        assert len(data.keys()) == 1
        klass, values = data.popitem()
        try:
            import damast
            datarange_subclass = getattr(damast.core.data_description, klass)
            return datarange_subclass.from_data(data=values, dtype=dtype)
        except AttributeError as e:
            msg = f"DataRange.from_dict: unknown range definition '{klass}' -- {e}"
            raise ValueError(msg) from e


class ListOfValues:
    """
    Represent a list of values.
    """

    #: Values in this list
    values: List[Any]

    def __init__(self, values: List[Any]):
        """
        Initialise ListOfValue

        :param values: values that define this list
        :raise ValueError: Raises if values is not a list.
        """
        if not isinstance(values, list):
            raise ValueError(
                f"{self.__class__.__name__}.__init__: required list of values for initialisation, got {type(values)}"
            )

        self.values = values

    def is_in_range(self, value) -> bool:
        """
        Check if value can be found in the list of values.

        :param value: Value to test
        :return: True if value is in the list, false otherwise
        """
        return value in self.values

    def is_not_in_range(self, value) -> bool:
        return not self.is_in_range(value)

    def __eq__(self, other) -> bool:
        """
        Check equality based on the 'values' property

        :param other: Other object
        :return: True if object are considered equal.
        """
        if self.__class__ != other.__class__:
            return False

        if self.values != other.values:
            return False

        return True

    @classmethod
    def from_data(cls, data: List[Any], dtype: Any) -> ListOfValues:
        """
        Create an instance from data and given datatype (dtype)

        :param data: The actual list of values
        :param dtype: Datatype of values in the list
        :return: instance
        :raise ValueError: If dtype is given, but does not match the value type in the given list of values
        """
        if data and dtype is not None:
            for sample in data:
                if sample is not None:
                    break

            actual_dtype = type(sample)
            if actual_dtype != dtype:
                raise ValueError(
                    f"{cls.__name__}.from_data: expected list of {dtype.__name__}, but received"
                    f" {actual_dtype}, {data=}"
                )
        return cls(values=data)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert to dictionary representing this object

        :return: dictionary
        """
        return dict(self)

    def __iter__(self):
        yield self.__class__.__name__, self.values


class MinMax(DataRange):
    """
    A Minimum-Maximum Range Definition

    :param min: minimum allowed value, data must be greater or equal
    :param max: maximum allowed value, data must be less or equal
    """

    #: Minimum / Lower Bound of range
    min: Any
    #: Maximum / Upper Bound of range
    max: Any

    #: Allow missing values for this data range
    allow_missing: bool

    def __init__(self, min: Any, max: Any, allow_missing: bool = True):
        """
        Constructor
        """
        super().__init__()

        if not min <= max:
            raise RuntimeError(f"DataRange.__init__: invalid range - min: {min} max: {max}")

        self.min = min
        self.max = max

        self.allow_missing = allow_missing

    def is_in_range(self, value: Any) -> bool:
        """
        Check if a value is in the defined range.

        :param value: data value

        :return: `True` if value is in the set range, `False` otherwise
        """
        # To make masked/missing values work
        try:
            if value is None:
                return self.allow_missing

            if np.isnan(value) or np.isnat(value):
                return self.allow_missing
        except Exception:
            pass

        return self.min <= value <= self.max

    @classmethod
    def from_data(cls, data: Dict[str, Any], dtype: Any) -> MinMax:
        """
        Load the MinMax range from the given dictionary specification

        .. highlight:: python
        .. code-block:: python

            min_max_range = MinMax.from_data({"min": 0, "max": 1}, dtype=float)

        :param data: The dictionary to describe the range
        :param dtype: value type that should be used to initialise this min max range
        :return: A :class:`MinMax` instance
        :raise KeyError: Raises if min/max keys are missing in the dictionary
        """
        for required_key in ["min", "max"]:
            if required_key not in data:
                raise KeyError(f"{cls.__name__}.from_data: missing '{required_key}'")

        min_value = data["min"]
        max_value = data["max"]

        if dtype is not None:
            min_value = DataElement.create(data["min"], dtype)
            max_value = DataElement.create(data["max"], dtype)

        keys = {
            'min': min_value,
            'max': max_value
        }

        if "allow_missing" in data:
            keys['allow_missing'] = data['allow_missing']

        return cls(**keys)

    def __eq__(self, other) -> bool:
        """
        Check equality based on min/max properties

        :param other: Other object
        :return: `True` if objects are considered the same, `False` otherwise
        """
        if self.__class__ != other.__class__:
            return False

        if self.min != other.min:
            return False

        if self.max != other.max:
            return False

        return True

    def __repr__(self) -> str:
        """
        Create representation for this object:

        :return: Representation for this instance
        """
        return f"{self.__class__.__name__}[{self.min}, {self.max}]"

    def merge(self, other: MinMax) -> MinMax:
        """
        Extend the range based on another range definition.

        If any of the DataRange instances does not allow missing value, the resulting instance
        will as not allow missing values.

        :param other: MinMax object to extend the bound of the current one
        """
        return MinMax(
                min=min(self.min, other.min),
                max=max(self.max, other.max),
                # Since allow missing is by default true - explicitly disabling it should
                # be propagated when merging
                allow_missing=self.allow_missing and other.allow_missing
        )

    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        Create a dictionary representing this object

        :return: dictionary
        """
        return dict(self)

    def __iter__(self):
        yield self.__class__.__name__, {"min": self.min, "max": self.max, "allow_missing": self.allow_missing}


class CyclicMinMax(MinMax):
    """
    Define a cyclic min max range
    """

class NumericValueStats(BaseModel):
    mean: float = 0.0
    stddev: float = 0.0
    total_count: int  = 0
    null_count: int = 0

    def merge(self, other: NumericValueStats) -> NumericValueStats:
        combined_mean = (
                    self.mean*self.total_count
                    + other.mean*other.total_count
                )/(
                    self.total_count
                    +other.total_count
                )
        a = (self.total_count-1)*self.stddev**2+(other.total_count-1)*other.stddev**2
        b = self.total_count*other.total_count/(self.total_count + other.total_count)
        c = b*(self.mean - other.mean)**2
        d = self.total_count + other.total_count -1
        combined_stddev = math.sqrt((a + c)/d)

        return NumericValueStats(
            mean=combined_mean,
            stddev=combined_stddev,
            total_count=self.total_count+other.total_count,
            null_count=self.null_count+other.null_count
        )





"""
Main argument parser and CLI entry point.
"""
import sys
import traceback as tb
import warnings
from argparse import ArgumentParser
from pathlib import Path

from damast.cli.base import BaseParser
from damast.cli.data_annotate import DataAnnotateParser
from damast.cli.data_converter import DataConvertParser
from damast.cli.data_inspect import DataInspectParser
from damast.cli.data_processing import DataProcessingParser
from damast.cli.experiment import ExperimentParser


class MainParser(ArgumentParser):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.description = "damast - data processing with annotation"

        self.add_argument("-w", "--workdir", default=str(Path(".").resolve()))
        self.add_argument("-v", "--verbose", action="store_true")
        self.add_argument("--loglevel", dest="loglevel", type=int, default=10, help="Set loglevel to display")
        self.add_argument("--logfile", dest="logfile", type=str, default=None,
                          help="Set file for saving log (default prints to terminal)")

        self.subparsers = self.add_subparsers(help='sub-command help')

    def attach_subcommand_parser(self,
                                 subcommand: str,
                                 help: str,
                                 parser_klass: BaseParser
                                 ):
        parser = self.subparsers.add_parser(subcommand, help=help)
        parser_klass(parser=parser)


def run():
    """
    Run the main command line interface
    """
    main_parser = MainParser()
    main_parser.attach_subcommand_parser(subcommand="annotate",
                                         help="Annotate a dataframe",
                                         parser_klass=DataAnnotateParser)

    convert_help_desc = "Convert a dataset (set of .csv-files) to a .h5-file (containing the data)" +\
        " and .yml-file (containing data specification)"
    main_parser.attach_subcommand_parser(subcommand="convert",
                                         help=convert_help_desc,
                                         parser_klass=DataConvertParser)

    experiment_help_desc = "Run a predefined experiment"
    main_parser.attach_subcommand_parser(subcommand="experiment",
                                         help=experiment_help_desc,
                                         parser_klass=ExperimentParser)

    inspect_help_desc = "Inspect a dataset given by filename or pattern and associated metadata"
    main_parser.attach_subcommand_parser(subcommand="inspect",
                                         help=inspect_help_desc,
                                         parser_klass=DataInspectParser)

    main_parser.attach_subcommand_parser(subcommand="process",
                                         help="Process data by running a predefined pipeline",
                                         parser_klass=DataProcessingParser)


    args = main_parser.parse_args()
    if hasattr(args, "active_subparser"):
        try:
            args.active_subparser.execute(args)
        except Exception as e:
            if args.verbose:
                tb.print_exception(e)
            else:
                print(f"\033[91mError: {e}\033[00m")
            sys.exit(1)
    else:
        main_parser.print_help()


if __name__ == "__main__":
    run()



"""
Module to define known constraints of the data and introduction of constants for column names and field values.
"""
__all__ = ["ColumnName",
           "FieldValue",
           "MMSI"
           ]

from typing import ClassVar


class CourseOverGround:
    min_value: ClassVar[float] = -360.0
    max_value: ClassVar[float] = 360.0

    unit: str = "deg"


class Heading:
    min_value: ClassVar[float] = 0.0
    max_value: ClassVar[float] = 359.0

    unit: str = "deg"


class Bearing:
    min_value: ClassVar[float] = 0.0
    max_value: ClassVar[float] = 359.0

    unit: str = "deg"


class SpeedOverGround:
    min_value: ClassVar[float] = 0.0
    max_value: ClassVar[float] = 102.0

    unit: str = "knots"


class RateOfTurn:
    min_value: ClassVar[float] = -720.0
    max_value: ClassVar[float] = 720.0

    unit: str = "deg"


class MMSI:
    """
    Class representing the MMSI data and corresponding (known) constraints
    """

    #: Minimum value for the MSSI which comes from a predefined allowed-range
    min_value: ClassVar[int] = 200000000

    #: Maximum value for the MSSI which comes from a predefined allowed-range
    max_value: ClassVar[int] = 799999999

    def __init__(self,
                 mmsi: int):
        self.value = mmsi

    @property
    def mid(self) -> int:
        return self.country_iso_code

    @property
    def country_iso_code(self) -> int:
        """
        Get the first three digits of the mmsi

        :return: Country code
        """
        return int(self.value / 1000000)

    @property
    def national_id(self) -> int:
        """
        Get the national id (i.e. 6 last digits of the mmsi)

        :return: National id
        """
        return self.value - self.country_iso_code * 1000000


class ColumnName:
    """
    Collection of existing column names to provide a single point of definition.
    """
    MMSI: str = "MMSI"
    COURSE_OVER_GROUND: str = "COG"
    SPEED_OVER_GROUND: str = "SOG"
    LATITUDE: str = "LAT"
    LONGITUDE: str = "LON"
    TIMESTAMP: str = "timestamp"
    VESSEL_TYPE: str = "vessel_type"
    VESSEL_TYPE_GFW: str = "vessel_class_gfw"
    STATUS: str = "Status"
    MESSAGE_TYPE: str = "MessageType"
    DATE_TIME_UTC: str = "date_time_utc"
    SOURCE: str = "source"

    # Computed Data
    DATA_MONTH: str = "data_month"
    FISHING_TYPE: str = "fishing_type"
    DISTANCE_CLOSEST_SATELLITE: str = "distance_closest_satellite"
    DISTANCE_CLOSEST_ANCHORAGE: str = "distance_closest_anchorages"
    HEADING: str = "Heading"

    TRAJECTORIES: str = "trajectories"
    COMBINATION: str = "combination"
    HISTORIC_SIZE: str = "HISTORIC_SIZE"
    HISTORIC_SIZE_REVERSE: str = "HISTORIC_SIZE_REVERSE"
    DELTA_DISTANCE: str = "DeltaDistance"
    LAST_POSITION: str = "last_position"


class FieldValue:
    VESSEL_TYPE_UNSPECIFIED: str = "unspecified"
    UNDEFINED: int = -1



import traceback as tc

from damast.core.dataframe import AnnotatedDataFrame
from damast.core.transformations import Transformer


class Pipeline:
    steps: list[Transformer]

    def __init__(self, steps: list[Transformer]):
        self.steps = steps

    def transform(self, df: AnnotatedDataFrame):
        return self.run(df=df)

    def run(self, df: AnnotatedDataFrame):
        for idx, step in enumerate(self.steps):
            try:
                df = step.fit_transform(df)
                AnnotatedDataFrame.ensure_type(df)
            except Exception as e:
                msg = ''.join(tc.format_exception(e)[-2:])
                raise RuntimeError(f"Step #{idx} in pipeline ({step.__class__.__name__}) failed: name_mappings: {step.name_mappings}\n\
                        {msg}")
        return df



import numpy as np
import polars

from damast.core.dataframe import AnnotatedDataFrame


class Transformer:
    def transform(self, df: AnnotatedDataFrame):
        return df

    def fit(self, df):
        pass

    def transform(self, df):
        return df

    def fit_transform(self, df: AnnotatedDataFrame):
        self.fit(df=df)
        return self.transform(df=df)

class CycleTransformer(Transformer):
    def __init__(self, features: list[str], n: int):
        self.features = features
        self.n = n

    def transform(self, df: AnnotatedDataFrame):
        if type(df) != AnnotatedDataFrame:
            raise ValueError(f"Transformer requires 'AnnotatedDataFrame',"
                    f" but got '{type(df)}")
        clone = df.copy()

        for feature in self.features:
            clone._dataframe = clone._dataframe.with_columns(
                    (np.cos(polars.col(feature)*2*np.pi) / self.n).alias(f"{feature}_x")
                )
            clone._dataframe = clone._dataframe.with_columns(
                    (np.cos(polars.col(feature)*2*np.pi) / self.n).alias(f"{feature}_y")
                )
        return clone



"""
Module to collect the classes to define meta data
"""
from __future__ import annotations

import ast
import builtins
import inspect
import logging
import os
import re
import traceback
import warnings
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import polars as pl
import yaml

from .annotations import Annotation, History
from .constants import (
    DAMAST_HDF5_COLUMNS,
    DAMAST_HDF5_ROOT,
    DAMAST_SPEC_SUFFIX,
    DAMAST_SUPPORTED_FILE_FORMATS,
    )
from .data_description import DataElement, DataRange, MinMax, NumericValueStats
from .formatting import DEFAULT_INDENT
from .types import DataFrame, XDataFrame
from .units import Unit, unit_registry, units

__all__ = [
    "ArtifactSpecification",
    "DataCategory",
    "DataSpecification",
    "MetaData",
    "Status",
    "ValidationMode",
]

logger = logging.getLogger(__name__)

class DataCategory(str, Enum):
    """
    Specification of what state a data-set is in (static or dynamic)
    """

    DYNAMIC = "dynamic"
    """Data-category is dynamic"""

    STATIC = "static"
    """Data-category is static"""


class Status(str, Enum):
    """
    Denoting if :class:`DataSpecification` adheres to :class:`DataSpecification.Fulfillment`
    """

    OK = "OK"
    """Data-specification is fulfilled"""

    FAIL = "FAIL"
    """Data-specification is not adhered to"""

class ArtifactSpecification:
    """
    Specification of a (file) artifact that might be generated during data processing.

    :param requirements: Dictionary describing expected path artifacts, key is a descriptor,
        value an absolute path or a relative path (pattern)
    """

    #: Dictionary mapping an artifact description to a file glob pattern
    artifacts: Dict[str, str]

    def __init__(self, requirements: Dict[str, Any]) -> None:
        self.artifacts = requirements

    def validate(self, base_dir: Path):
        """
        Validate the list of artifacts.

        :param base_dir: Relative paths/patterns will be interpreted with ``base_dir`` as current working directory
        :raise RuntimeError: When a matching file could not be found
        """
        for name, path_pattern in self.artifacts.items():
            a_path = Path(path_pattern)
            if a_path.is_absolute():
                if not a_path.exists():
                    raise RuntimeError(
                        f"{self.__class__.__name__}.validate: artifact '{a_path}' is "
                        "missing"
                    )
            else:
                files = list(Path(base_dir).glob(path_pattern))
                if not files:
                    raise RuntimeError(
                        f"{self.__class__.__name__}.validate: no artifact matching "
                        f" {path_pattern} found in '{base_dir}'"
                    )


class ValidationMode(str, Enum):
    """
    Specification of how to apply meta-data to :class:`DataFrame`
    """

    IGNORE = "IGNORE"
    """Metadata will not be validated"""

    READONLY = "READONLY"
    """Metadata cannot be changed"""

    UPDATE_DATA = "UPDATE_DATA"
    """Data should be updated to comply with the metadata"""

    UPDATE_METADATA = "UPDATE_METADATA"
    """Metadata should be updated to comply with the data"""


class DataSpecification:
    """
    Specification of a single column and/or dimension in a :class:`damast.core.dataframe.DataFrame`.

    :param name: Name of the element
    :param category: A :class:`DataCategory` element
    :param is_optional: If column is optional
    :param abbreviation: Abbreviation of name
    :param representation_type: The underlying representation type for this data element
    :param missing_value: The representation of a missing value for this data element
    :param unit: The (physical) unit of this data element
    :param precision: The precision of this data element
    :param value_range: The allowed data range, which remains as ``None`` if unrestricted
    :param value_stats: The statistics about the data
    :param value_meanings: A description of value meanings
    """

    class Key(str, Enum):
        """
        Valid inputs to a :class:`DataSpecification`
        """

        name = "name"
        description = "description"
        category = "category"
        is_optional = "is_optional"
        abbreviation = "abbreviation"

        representation_type = "representation_type"
        missing_value = "missing_value"
        precision = "precision"
        unit = "unit"
        value_range = "value_range"
        value_stats = "value_stats"
        value_meanings = "value_meanings"

    class Fulfillment:
        """
        Base-class used for specific fulfillments
        """

        # Would use Key here, but the concept for that is not yet ironed out in Python
        status: Dict[Enum, Dict[str, Union[Status, str]]]

        def __init__(self):
            self.status = {}

        def is_met(self) -> bool:
            """
            Check the status property to see if the fulfillment is considered as met.

            :raises KeyError: If `self.status` does not have the key ``"status"``
            :return: ``True`` if fulfillment is met, else ``False``.
            """
            for _, v in self.status.items():
                if "status" not in v:
                    raise KeyError(
                        f"{self.__class__.__name__}: internal error. Status misses key 'status'"
                    )

                if v["status"] == Status.FAIL:
                    return False
            return True

        def __repr__(self):
            status = {}
            for k, v in self.status.items():
                msg = v["status"].value
                if "message" in v:
                    msg += f" {v['message']}"
                status[k.value] = msg

            return f"{self.__class__.__name__}[{status}]"

    class MissingColumn(Fulfillment):
        """
        Represent a missing column fulfillment.

        :param missing_column: Name of the missing column
        :param known_columns: Names of the (by specification) known columns
        """

        missing_column: str
        known_columns: List[str]

        def __init__(self, missing_column: str, known_columns: List[str]):
            super().__init__()

            self.missing_column = missing_column
            self.known_columns = known_columns

        def is_met(self) -> bool:
            """
            Check if fulfillment is met.

            :return: Always returns False, since this represents a missing fulfillment
            """
            return False

        def __repr__(self) -> str:
            return f"{self.__class__.__name__}: '{self.missing_column}', known: {','.join(self.known_columns)}"

    #: Name associated
    name: str
    #: Description of the data
    description: Optional[str] = None
    #: Category of data
    category: Optional[DataCategory] = None
    #: Whether this data element needs to be present
    is_optional: Optional[bool] = None
    abbreviation: Optional[str] = None

    #: The underlying representation type for this data element
    representation_type: Any = None
    #: The representation of a missing value for this data element
    missing_value: Any = None

    #: The unit of this data element
    _unit: Optional[Unit] = None
    #: The precision of this data element
    # FIXME: The input to precision could be a `List[float]`, but this is not currently handled
    precision: Optional[float] = None

    #: The allowed data range, which remains None when being unrestricted
    value_range: Optional[DataRange] = None
    #: An explanation - str-based descriptor for the range, if this is a dictionary then it must provide
    #: a mapping from the value to a human-readable descriptor
    value_meanings: Optional[Dict[Any, str]] = None

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        category: Optional[Union[str, DataCategory]] = None,
        is_optional: bool = False,
        abbreviation: Optional[str] = None,
        representation_type: Any = None,
        missing_value: Any = None,
        unit: Optional[Unit] = None,
        precision: Any = None,
        value_range: Optional[DataRange] = None,
        value_stats: Optional[NumericValueStats] = None,
        value_meanings: Optional[Dict[Any, str]] = None,
    ):
        """
        Constructor
        """

        if name is None:
            raise ValueError(
                f"{self.__class__.__name__}.__init__: " f"name cannot be None"
            )

        self.name = name
        if isinstance(category, str):
            self.category = DataCategory(category)
        else:
            self.category = category

        self.is_optional = is_optional
        if abbreviation is not None and abbreviation != "":
            self.abbreviation = abbreviation

        self.representation_type = representation_type
        self.missing_value = missing_value

        self.unit = unit
        self.precision = precision
        self.value_range = value_range
        self.value_stats = value_stats
        self.value_meanings = value_meanings
        self.description = description
        self._validate()

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        return all(
            getattr(self, m.value) == getattr(other, m.value)
            for m in DataSpecification.Key
        )

    def __repr__(self):
        return f"{self.__class__.__name__}[{dict(self)}]"

    def _validate(self):
        """
        Validate the data specification

        :raises ValueError: If the specification is invalid
        """
        if self.value_meanings and not self.value_range:
            raise ValueError(
                f"{self.__class__.__name__}.__init__: "
                f" value meanings requires value range to be set"
            )

        if self.value_meanings:
            for k, _ in self.value_meanings.items():
                if k not in self.value_range:
                    raise ValueError(
                        f"{self.__class__.__name__}.validate: "
                        " value {k} is not in the defined value range"
                    )

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value: Unit | str | None):
        if isinstance(value, str):
            # This will include astropy and custom (damast) units
            self._unit = Unit(value)
        elif isinstance(value, units.UnitBase):
            self._unit = value
        elif value is None:
            pass
        else:
            raise ValueError(f"MetaData: cannot set unit using {value=}")

    @classmethod
    def resolve_representation_type(cls, type_name: str) -> Any:
        """
        Resolve the representation type from a given string.

        .. highlight:: python
        .. code-block:: python

            int_type = DataSpecification.resolve_representation_type("int")
            assert int_type == int

        :param type_name: Name of the type
        :return: Instance of the type object
        :raise ValueError: Raises if ``type_name`` cannot be resolved to a known type (in builtins or :mod:`polars`)
        """
        exceptions: List[Any] = []
        try:
            if type_name.lower() in ["string", "str"]:
                dtype = str
            else:
                dtype = getattr(builtins, type_name)
            return dtype
        except AttributeError as e:
            exceptions.append(e)

        try:
            m = re.match("(.*)\((.*)\)", type_name)
            call_args = None
            if m is not None:
                type_name = m.group(1)
                call_args = m.group(2)
            dtype = getattr(pl.datatypes, type_name)
            if call_args:
                dtype = eval(f"pl.datatypes.{dtype}({call_args})")
            return dtype
        except TypeError as e:
            exceptions.append(e)

        raise ValueError(
            f"{cls.__name__}.resolve_representation_type: "
            f" could not find type '{type_name}'."
            f"It does not exist in builtins, nor "
            f" is this a known polar datatype -- {exceptions}"
        )

    def __iter__(self):
        yield "name", self.name
        if self.is_optional is not None:
            yield "is_optional", self.is_optional

        if self.abbreviation is not None:
            yield "abbreviation", self.abbreviation

        if self.category is not None:
            yield "category", self.category.value

        if self.representation_type is not None:
            if inspect.isclass(self.representation_type):
                yield "representation_type", self.representation_type.__name__
            elif isinstance(self.representation_type, pl.datatypes.DataType):
                yield "representation_type", str(self.representation_type)
            else:
                raise TypeError(
                    f"{self.__class__.__name__}.__iter__ failed to identify representation_type from"
                    f" {self.representation_type}"
                )

        if self.missing_value is not None:
            yield "missing_value", self.missing_value

        if self.unit is not None:
            yield "unit", self.unit.to_string()

        if self.precision is not None:
            yield "precision", self.precision

        if self.value_range is not None:
            yield "value_range", dict(self.value_range)

        if self.value_stats is not None:
            yield "value_stats", self.value_stats.__dict__

        if self.value_meanings is not None:
            yield "value_meanings", self.value_meanings

        if self.description is not None:
            yield "description", self.description

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> DataSpecification:
        """
        Load the data specification from a given dictionary.

        This function in intended to deserialize specifications from a string-based
        representation

        :param data: data dictionary using primitive data types to represent the specification
        :return: the loaded data specification
        :raise KeyError: Raise if a required key is missing
        """
        for key in [cls.Key.name, cls.Key.is_optional]:
            if key.value not in data:
                raise KeyError(f"{cls.__name__}.from_dict: missing '{key.value}'")

        abbreviation = None
        if cls.Key.abbreviation.value in data:
            abbreviation = data[cls.Key.abbreviation.value]
            if not isinstance(abbreviation, str):
                raise TypeError(
                    f"{cls.__name__}.from_dict: " f"abbreviation must be of type 'str'"
                    f", but was {type(abbreviation)} - {abbreviation}"
                )

        data_category = None
        if cls.Key.category.value in data:
            data_category = data[cls.Key.category]
            if not isinstance(data_category, DataCategory):
                data_category = DataCategory(data_category)

        spec = cls(
            name=data[cls.Key.name],
            category=data_category,
            is_optional=bool(data[cls.Key.is_optional]),
            abbreviation=abbreviation,
        )

        if cls.Key.description in data:
            spec.description = data[cls.Key.description]

        if cls.Key.representation_type.value in data:
            spec.representation_type = cls.resolve_representation_type(
                data[cls.Key.representation_type.value]
            )
            if cls.Key.missing_value.value in data:
                spec.missing_value = DataElement.create(
                    data[cls.Key.missing_value.value], spec.representation_type
                )
            if cls.Key.precision.value in data:
                spec.precision = DataElement.create(
                    data[cls.Key.precision.value], spec.representation_type
                )
        # else: Missing representation type, missing value and precision will not be loaded

        if cls.Key.unit.value in data:
            # Check if unit is part of the default definitions
            unit_value = data[cls.Key.unit.value]
            unit = None
            if isinstance(unit_value, str):
                # This will include astropy and custom (damast) units
                unit = Unit(unit_value)
            elif isinstance(unit_value, Unit):
                unit = unit_value
            else:
                raise RuntimeError(
                    f"{cls.__name__}.from_dict: cannot interpret unit type "
                    f"'{type(unit_value)}"
                )

            spec.unit = unit

        if cls.Key.value_range.value in data:
            value_range = data[cls.Key.value_range.value]
            if isinstance(value_range, DataRange):
                spec.value_range = value_range
            else:
                if isinstance(value_range, str):
                    value_range = ast.literal_eval(value_range)
                spec.value_range = DataRange.from_dict(
                    value_range, dtype=spec.representation_type
                )

            if cls.Key.value_meanings.value in data:
                spec.value_meanings = data[cls.Key.value_meanings.value]

            if cls.Key.value_stats.value in data:
                if not str(spec.representation_type).lower().startswith("str"):
                    spec.value_stats = NumericValueStats(**data[cls.Key.value_stats.value])

        return spec

    @classmethod
    def to_str(cls, specs: List[DataSpecification], indent_level=0):
        """
        Generate string representation for list of specs

        :param specs: List of data specifications
        :param indent_level: indentation levels
        :param spaces: number of space for one indentation level
        :return: String representation of specification
        """
        hspace = DEFAULT_INDENT * indent_level

        data = ""
        for spec in specs:
            data += hspace + spec.name
            if spec.unit is not None:
                data += f"[unit: {spec.unit.to_string()}]"
            data += "\n"
        return data

    def apply(
        self,
        df: DataFrame | XDataFrame,
        column_name: str,
        validation_mode: ValidationMode
        ) -> DataFrame:
        """
        Apply the metadata object to the dataframe

        :param df: Dataframe that should be associated with the data-specification
        :param column_name: Name of the column that need to be validated/checked
        :param validation_mode: Mode which shall apply to updating either data or
            metadata, when encountering inconsistencies
        """

        df_type = type(df)
        if df_type == pl.LazyFrame:
            pass
        elif df_type == pl.DataFrame:
            df = df.lazy()
        elif df_type == XDataFrame:
            df = df._dataframe
        else:
            raise TypeError("MetaData.apply: dataframe must be either "
                " polars.LazyFrame,"
                " polars.DataFrame,"
                " or damast.core.types.XDataFrame"
            )

        if validation_mode == ValidationMode.IGNORE:
            return df

        # Check if representation type is the same and apply known metadata
        if validation_mode == ValidationMode.READONLY:
            xdf = XDataFrame(df)
            if self.representation_type is not None:
                dtype = xdf.dtype(column_name)
                if dtype.from_python(self.representation_type) != self.representation_type and \
                    dtype.to_python() != self.representation_type:
                    raise ValueError(
                        f"{self.__class__.__name__}.apply: column '{column_name}':"
                        f" expected representation type: {self.representation_type},"
                        f" but got '{dtype}'"
                    )

            if self.value_range:
                min_value, max_value = xdf.minmax(column_name)
                if not self.value_range.is_in_range(min_value):
                    raise ValueError(
                        f"{self.__class__.__name__}.apply: minimum value '{min_value}'"
                        f" lies outside of range {self.value_range} for column '{column_name}'"
                    )
                if not self.value_range.is_in_range(max_value):
                    raise ValueError(
                        f"{self.__class__.__name__}.apply: maximum value '{max_value}'"
                        f" lies outside of range {self.value_range} for column '{column_name}'"
                    )
            return df

        if validation_mode == ValidationMode.UPDATE_DATA:
            xdf = XDataFrame(df)
            if self.representation_type is not None:
                dtype = xdf.dtype(column_name)
                if dtype != self.representation_type:
                    warnings.warn(
                        f"{self.__class__.__name__}.apply: column '{column_name}':"
                        f" expected representation type: {self.representation_type},"
                        f" but got '{dtype}'"
                    )
                    xdf.set_dtype(column_name, self.representation_type)

            if self.value_range:
                if self.missing_value is None:
                    warnings.warn(
                        f"Filtering out for column '{column_name}' values that are out of range."
                    )
                    xdf._dataframe = xdf._dataframe.filter(
                            (pl.col(column_name) >= self.value_range.min) &
                            (pl.col(column_name) <= self.value_range.max)
                        )
                else:
                    xdf._dataframe = xdf._dataframe.with_columns(
                                pl.when(
                                    (pl.col(column_name) < self.value_range.min) |
                                    (pl.col(column_name) > self.value_range.max)
                                ).then(self.missing_value)
                                .otherwise(pl.col(column_name))
                                .alias(column_name)
                              )
            return xdf._dataframe

        if validation_mode == ValidationMode.UPDATE_METADATA:
            xdf = XDataFrame(df)
            self.representation_type = xdf.dtype(column_name)

            try:
                min_value, max_value = xdf.minmax(column_name)
                if self.value_range:
                    if isinstance(self.value_range, MinMax):
                        self.value_range = self.value_range.merge(MinMax(min_value, max_value))
                else:
                     warnings.warn(
                         f"Setting MinMax range ({min_value}, {max_value}) for {column_name}"
                     )
                     self.value_range = MinMax(min_value, max_value)
            except ValueError:
                # Type might not be numeric
                pass
            return xdf._dataframe

    def get_fulfillment(self, data_spec: DataSpecification) -> Fulfillment:
        """
        Check the fulfillment of an existing and expected data specification against an available one.

        The current instance used for calling this method sets the expectations.

        :param data_spec: The available data specification
        :return: The degree of fulfillment
        """
        fulfillment = DataSpecification.Fulfillment()

        for key in DataSpecification.Key:
            expected_value = getattr(self, key.value)
            # Nothing expected so status is OK
            if expected_value is None:
                fulfillment.status[key] = {"status": Status.OK}
                continue

            # Value expected, but no value available in the tested spec so FAIL
            spec_value = getattr(data_spec, key.name)
            if spec_value is None:
                fulfillment.status[key] = {
                    "status": Status.FAIL,
                    "message": "column has no precision defined",
                }

            # some special handling is required for individual keys
            if key == DataSpecification.Key.precision:
                    if expected_value < spec_value:
                        fulfillment.status[key] = {
                            "status": Status.FAIL,
                            "message": "'data has insufficient precision: "
                            f" required '{expected_value}',"
                            f" available '{spec_value}'",
                        }
                    else:
                        fulfillment.status[key] = {"status": Status.OK}
            elif key == DataSpecification.Key.representation_type:
                if hasattr(spec_value, "to_python"):
                    spec_value = spec_value.to_python()

                if not issubclass(spec_value, expected_value):
                    fulfillment.status[key] = {
                        "status": Status.FAIL,
                        "message": "'representation type is not a subclass of the expected:"
                        f" required '{expected_value}',"
                        f" available '{spec_value}'",
                    }
                else:
                    fulfillment.status[key] = {"status": Status.OK}
            else:
                expected_value = getattr(self, key.value)
                if expected_value is not None:
                    spec_value = getattr(data_spec, key.name)
                    if spec_value is None:
                        fulfillment.status[key] = {
                            "status": Status.FAIL,
                            "message": f"Expected '{spec_value=}' == '{expected_value=}'",
                        }
                        continue

                    if expected_value != spec_value:
                        fulfillment.status[key] = {
                            "status": Status.FAIL,
                            "message": f"Expected '{spec_value=}' == '{expected_value=}'",
                        }
                        continue

                    fulfillment.status[key] = {"status": Status.OK}
        return fulfillment

    @classmethod
    def from_requirements(cls, requirements: Dict[str, Any]) -> List[DataSpecification]:
        """
        Get the list of DataSpecification from dictionary (keyword argument) based representation.

        DataSpecification.from_requirements(requirements={ "column_name": { "unit": units.m }})

        :param requirements: List of dictionaries describing the data specification
        :return: List of expectations
        """
        required_specs = []
        for k, v in requirements.items():
            kwargs = v
            # consider the key to be the column name
            kwargs[DataSpecification.Key.name.value] = k
            if DataSpecification.Key.category.value not in kwargs:
                kwargs[DataSpecification.Key.category.value] = None

            required_spec = cls(**kwargs)
            required_specs.append(required_spec)
        return required_specs

    def merge(self, other: DataSpecification) -> DataSpecification:
        """Merge the current data-specification and one input specification
        into a single :class:`DataSpecification` object.

        :param other: The other data-specification
        :raises ValueError: If :attr:`DataSpecification.name` differs between the two specs.
        :raises ValueError: If the data-specifications have overlapping attributes, that have distinct non-``None``
            values, the function throws an error.
        """
        if self.name != other.name:
            raise ValueError(
                f"{self.__class__.__name__}.merge: cannot merge specs with different name property"
            )

        ds = DataSpecification(name=self.name)
        for key in self.Key:
            if key == self.Key.name:
                continue

            this_value = getattr(self, key.value)
            other_value = getattr(other, key.value)

            if key == self.Key.representation_type:
                if hasattr(this_value, "to_python"):
                    this_value = this_value.to_python()
                if hasattr(other_value, "to_python"):
                    other_value = other_value.to_python()

            if this_value is None:
                setattr(ds, key.value, other_value)
            elif other_value is None:
                setattr(ds, key.value, this_value)
            elif this_value == other_value:
                setattr(ds, key.value, this_value)
            else:
                if hasattr(this_value, "merge"):
                    merged_value = this_value.merge(other_value)
                    setattr(ds, key.value, merged_value)
                else:
                    raise ValueError(
                        f"{self.__class__.__name__}.merge cannot merge specs for '{self.name}': value for '{key.value}' differs: "

                        f" on self: '{this_value}' vs. other: '{other_value}'"
                    )
        return ds

    @classmethod
    def merge_lists(
        cls, a_specs: List[DataSpecification], b_specs: List[DataSpecification]
    ) -> List[DataSpecification]:
        """
        Merge two lists of data-specifications into a single list.

        If a :class:`DataSpecification` in ``a_specs`` has the same :attr:`DataSpecification.name`
        as one in ``b_specs``, the specs are merged using :func:`merge`.

        :param a_specs: First list of specs
        :param b_specs: Second list of specs
        """
        result_specs: List[DataSpecification] = []

        b_column_dict = {x.name: x for x in b_specs}
        a_columns_names = []

        for a_spec in a_specs:
            column_name = a_spec.name
            if column_name in b_column_dict:
                # Need to check merge
                b_column_spec = b_column_dict[column_name]

                merged_spec = a_spec.merge(other=b_column_spec)
                result_specs.append(merged_spec)
            else:
                result_specs.append(a_spec)

            a_columns_names.append(a_spec.name)

        for b_spec in b_specs:
            if b_spec.name not in a_columns_names:
                result_specs.append(b_spec)

        return result_specs


class MetaData:
    """
    The representation for metadata that can be associated with a :class:`polars.LazyFrame`.

    :param columns: (Ordered) list of column specifications
    :param annotations:  List of annotations for this dataframe.

    .. note::
        Each annotation is assumed to have a unique :attr:`Annotation.name`.
    """

    class Key(str, Enum):
        """
        The two allowed objects in the :class:`MetaData` and their string representation.
        This is used for the dictionary representation dict(<MetaData>), :func:`MetaData.__iter__`,
        :func:`MetaData.from_dict`.
        """

        columns = "columns"
        annotations = "annotations"

    class Fulfillment:
        """
        A class containing all fulfillments (set of :class:`DataSpecification.Fulfillment`).

        A fulfillment describes whether a constraint on a column holds or not, added to :class:`MetaData`
        """

        column_fulfillments: Dict[str, DataSpecification.Fulfillment]

        def __init__(self):
            self.column_fulfillments = {}

        def is_met(self) -> bool:
            """Check if all fulfillments in the set of data-specifications are met."""
            return all(v.is_met() for k, v in self.column_fulfillments.items())

        def add_fulfillment(
            self, column_name: str, fulfillment: DataSpecification.Fulfillment
        ) -> None:
            """Add fulfillment to a given column.

            .. note::
                This function replaces any old fulfillment set on this column.
                Use :func:`DataSpecification.merge` to combine the existing spec
                and new spec prior to calling this function

            :param column_name: Name of column
            :param fulfillment: Fulfillment object
            """
            assert isinstance(fulfillment, DataSpecification.Fulfillment)
            self.column_fulfillments[column_name] = fulfillment

        def __repr__(self) -> str:
            txt = ""
            for column_name, fulfillment in self.column_fulfillments.items():
                txt += f"[{column_name}: {fulfillment}]"
            return txt

    #: Specification of columns in the
    columns: List[DataSpecification]

    # We store the annotations as a dictionary for easy lookup.
    # even if it means duplicating the name of the annotation as a key
    #: Dictionary containing all annotations
    _annotations: Dict[str, Annotation]

    def __init__(
        self,
        columns: List[DataSpecification],
        annotations: Optional[List[Annotation]] = None,
    ):
        assert isinstance(columns, list)
        self.columns = sorted(columns, key=lambda x: x.name)
        if annotations is None:
            self._annotations = {}
        else:
            if not isinstance(annotations, list):
                raise ValueError(f"Input annotations need to be a list, but was {annotations=}")

            unique_annotations = {annotation.name for annotation in annotations}
            if len(unique_annotations) != len(annotations):
                raise ValueError(
                    f"{self.__class__.__name__}: Set of annotations in metadata has duplicate names "
                    + f"got {[dict(an) for an in annotations]}"
                )
            self._annotations = {an.name: an for an in annotations}

    def add_annotation(self, annotation: Annotation):
        if annotation.name in self._annotations:
            raise KeyError(f"Annotation {annotation.name} already exists")

        self._annotations[annotation.name] = annotation

    def set_annotation(self, annotation: Annotation):
        self._annotations[annotation.name] = annotation

    @property
    def annotations(self) -> Dict[str, Annotation]:
        """Get dictionary of annotations"""
        return self._annotations

    def __eq__(self, other) -> bool:
        """
        Check if the current meta-data object is equal to another.

        .. note::
            This function considers two :class:`MetaData` objects
            with the same columns and annotations (but in different order)
            as different objects
        """
        if self.__class__ != other.__class__:
            return False

        if self.columns != other.columns:
            return False

        if len(self.annotations) != len(other.annotations):
            return False

        for k,v in self.annotations.items():
            if k in other.annotations and other.annotations[k] == v:
                continue
            return False

        return True

    def __iter__(self):
        columns = [dict(ds) for ds in self.columns]
        yield self.Key.columns.value, columns
        annotations = {}
        for key, value in self.annotations.items():
            assert key == value.name
            annotations[key] = dict(value)[key]
        yield "annotations", annotations

    def to_str(self, columns: list[str] | None = None, indent: int = 0, default_indent: str = " " * 4) -> str:
        hspace = " " * indent
        txt_repr = [f"{hspace}Annotations:"]
        for name, annotation in self.annotations.items():
            txt_repr.append(hspace + default_indent + f"{name}: {annotation.value}")

        for spec in self.columns:
            spec_dict = dict(spec)
            if columns and spec_dict['name'] not in columns:
                continue

            txt_repr.append(hspace + default_indent + f"{spec_dict['name']}:")
            for field_name, value in spec_dict.items():
                if field_name == "name":
                    continue
                txt_repr.append(
                    hspace + default_indent + default_indent + f"{field_name}: {value}"
                )

        return "\n".join(txt_repr)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> MetaData:
        """
        Create a :class:`MetaData` object by parsing a dictionary

        .. note::
            This function ignores all keys but those defined in :attr:`MetaData.Key`.
            These keys are required to generate the object from a dictionary.

        The items corresponding to the key :attr:`MetaData.Key.columns`
        should be a list of :class:`DataSpecification` compatible dictionaries.
        See :func:`DataSpecification.from_dict` for more information about those requirements.

        The items corresponding to :attr:`MetaData.Key.annotations` should be a dictionary of
        either :class:`damast.core.Annotation` compatible dictionaries (see :func:`damast.core.Annotation.from_dict`)
        or :class:`damast.core.History` compatible dictionaries (see :func:`damast.core.History.from_dict`)

        :param data: Input dictionary
        :raises KeyError: If missing required keys
        :raises RuntimeError: If data-specification dictionaries not compatible
        """
        for key in [cls.Key.annotations.value, cls.Key.columns.value]:
            if key not in data:
                raise KeyError(
                    f"{cls.__name__}.from_dict: Missing {key} key in MetaData"
                )

        data_specs = []
        for data_spec_dict in data[cls.Key.columns.value]:
            try:
                data_specification = DataSpecification.from_dict(data=data_spec_dict)
                data_specs.append(data_specification)
            except Exception as e:
                traceback.print_tb(e.__traceback__)
                raise RuntimeError(
                    f"{cls.__name__}.from_dict: could not"
                    f" process column specification: {data_spec_dict} -- {e}"
                ) from e

        annotations = []
        for annotation_key, annotation_value in data[cls.Key.annotations.value].items():
            annotation = None
            if annotation_key == Annotation.Key.History.value:
                annotation = History.from_dict(data={annotation_key: annotation_value})
            else:
                annotation = Annotation.from_dict(
                    data={annotation_key: annotation_value}
                )
            annotations.append(annotation)

        return cls(columns=data_specs, annotations=annotations)

    @classmethod
    def load_yaml(cls, filename: Union[str, Path]) -> MetaData:
        """
        Load history from a `yaml` file.

        .. note::
            It is assumed that the input file follows a structure that is read in as
            a dictionary compatible with :func:`from_dict`.

        :param filename: The input file
        :raises FileNotFoundError: If file does not exist
        :return: A MetaData object
        """
        yaml_path = Path(filename)
        if not yaml_path.exists():
            raise FileNotFoundError(
                f"{cls.__name__}.load_yaml: file " f" '{filename}' does not exist"
            )

        with open(yaml_path, "r") as f:
            md_dict = yaml.load(f, Loader=yaml.SafeLoader)

        return cls.from_dict(data=md_dict)

    def save_yaml(self, filename: Union[str, Path]):
        """
        Save the current object into a file.

        :param filename: Filename to use for saving
        """
        with open(filename, "w") as f:
            # Do not sort the keys, but keep the entry order the way
            # the dictionary has been constructed
            yaml.dump(dict(self), f, sort_keys=False)

    def append_to_hdf(self, path: str | Path, overwrite: bool = False):
        # Add metadata
        import tables
        with tables.open_file(path, "a") as f:
            main_node = f.get_node(DAMAST_HDF5_ROOT)

            # Add annotations to main group
            for key, annotation in self.annotations.items():
                if (
                    key in main_node._v_attrs._f_list()
                    and main_node._v_attrs[key] != annotation
                ) and not overwrite:
                    raise RuntimeError(
                        f"{self.__class__.__name__}.save:"
                        f" attribute '{key}' present"
                        f" in file but different from user-defined"
                    )
                main_node._v_attrs[key] = annotation

            if DAMAST_HDF5_COLUMNS not in f:
                f.create_group(DAMAST_HDF5_ROOT, DAMAST_HDF5_COLUMNS.replace(DAMAST_HDF5_ROOT + "/",""))

            # Add attributes for columns
            for column_spec in self.columns:
                group_name = f"{DAMAST_HDF5_COLUMNS}/{column_spec.name}"
                if group_name not in f:
                    group = f.create_group(DAMAST_HDF5_COLUMNS, column_spec.name)
                else:
                    group = f.get_node(group_name)

                for key, value in dict(column_spec).items():
                    if isinstance(value, dict):
                        group._v_attrs[key] = str(value)
                    else:
                        group._v_attrs[key] = value


    def apply(
        self,
        df: pl.LazyFrame,
        validation_mode: ValidationMode = ValidationMode.READONLY,
    ) -> pl.LazyFrame:
        """Check that each column in the :class:`polars.LazyFrame` fulfills the
        data-specifications.

        :param df: The dataframe
        :param validation_mode: If a column does not comply to the spec, force the spec, e.g.,
            for a given range specification, map data within range.
        :raises: ValueError: If data-frame is missing a column in the Data-specification
        """
        assert isinstance(df, pl.LazyFrame), f"Expected polars.LazyFrame, got {type(df)}"

        for column_spec in self.columns:
            columns = df.compat.column_names
            if column_spec.name in columns:
                df = column_spec.apply(
                    df=df, column_name=column_spec.name, validation_mode=validation_mode
                )
            else:
                raise ValueError(
                    f"{self.__class__.__name__}.apply: missing column '{column_spec.name}' in dataframe."
                    f" Found {len(columns)} column(s): {','.join(columns)}"
                )
        return df

    def __contains__(self, column_name: str):
        """"""
        return any(colum_spec.name == column_name for colum_spec in self.columns)

    def drop(self, columns: str | list[str]):
        """
        Drop specification by column name
        """
        columns = [columns] if type(columns) == str else columns
        self.columns = [x for x in self.columns if x.name not in columns]

    def __getitem__(self, column_name: str) -> DataSpecification:
        """
        Get a data-specification by its column name
        """
        for column_spec in self.columns:
            if column_spec.name == column_name:
                return column_spec

        raise KeyError(
            f"{self.__class__.__name__}.__getitem__: failed to find column by name '{column_name}'"
        )

    def get_fulfillment(self, expected_specs: List[DataSpecification]) -> Fulfillment:
        """
        Get the fulfillment of the metadata with represent to the given data specification

        :param expected_specs: A list of data specifications
        :return: The fulfillment object
        """
        md_fulfillment = MetaData.Fulfillment()

        for expected_spec in expected_specs:
            if expected_spec.name in self:
                fulfillment = expected_spec.get_fulfillment(self[expected_spec.name])
                md_fulfillment.add_fulfillment(
                    column_name=expected_spec.name, fulfillment=fulfillment
                )
            else:
                md_fulfillment.add_fulfillment(
                    column_name=expected_spec.name,
                    fulfillment=DataSpecification.MissingColumn(
                        missing_column=expected_spec.name,
                        known_columns=[x.name for x in self.columns],
                    ),
                )
        return md_fulfillment

    def merge(self, other: MetaData) -> MetaData:
        self.columns = DataSpecification.merge_lists(self.columns , other.columns)
        return self

    @classmethod
    def search(cls, files: list[str | Path]) -> tuple[MetaData | None, str | None]:
        """
        Search for the metadata specfile for a given list of files
        """
        logger.info("Metadata is required, so searching now for an existing annotation file")
        try:
            commonpath = os.path.commonpath(files)
        except Exception as e:
            logger.debug(e)
            return None, None, None

        if len(files) == 1:
            commonpath = Path(commonpath).parent

        commonprefix = os.path.commonprefix([Path(x).stem for x in files])
        metadata_file_candidates = [x for x in Path(commonpath).glob(f"{commonprefix}*{DAMAST_SPEC_SUFFIX}")]
        for f in metadata_file_candidates:
            try:
                return MetaData.load_yaml(filename=f), f, metadata_file_candidates
            except Exception as e:
                logger.debug(f"Loading {f} as metadata file failed -- {e}")

        return None, None, metadata_file_candidates

    @classmethod
    def specfile(cls, files: list[str | Path]) -> str:
        """
        Create the default filename for the spec file
        """
        if len(files) == 1:
            return Path(files[0]).with_suffix(DAMAST_SPEC_SUFFIX)

        commonpath = os.path.commonpath(files)
        commonprefix = os.path.commonprefix([Path(x).stem for x in files])
        return Path(commonpath) / f"{commonprefix}.collection{DAMAST_SPEC_SUFFIX}"

    def merge(self, other: Metadata) -> Metadata:
        column_specs = DataSpecification.merge_lists(self.columns, other.columns)
        annotations = []
        for k,v in self.annotations.items():
            try:
                values = v.value
                if v.value == other.annotations[k].value:
                    pass
                elif type(v.value) == type(other.annotations[k].value):
                    if type(v.value) == str:
                        values = [v.value, other.annotations[k].value]
                    elif type(v.value) == list:
                        values = v.value + other.annotations[k].value
                    elif type(v.value) == dict:
                        values = v.value.update(other.annotations[k].value)

                elif type(v.value) == list:
                    values = v.value + [other.annotations[k].value]
                elif type(other.annotations[k].value) == list:
                    values = [v.value] + other.annotations[k].value

                annotations.append(Annotation(k, values))
            except Exception as e:
                logger.warning(e)

        for k,v in other.annotations.items():
            if k not in self.annotations:
                annotations.append(v)

        return MetaData(column_specs, annotations)



import logging
import shutil
import subprocess
import sys
import tempfile
import time
import warnings
from pathlib import Path
from typing import Callable, ClassVar

DAMAST_ARCHIVE_SUPPORT_AVAILABLE = False
try:
    from ratarmountcore.compressions import ARCHIVE_FORMATS, COMPRESSION_FORMATS
    DAMAST_ARCHIVE_SUPPORT_AVAILABLE = True
except Exception as e:
    warnings.warn("ratarmount could not be loaded: archive support is not available")


from damast.core.constants import DAMAST_MOUNT_PREFIX
from damast.core.dataframe import AnnotatedDataFrame

logger = logging.getLogger(__name__)

class Archive:
    """
    Class to wrap and extract archive objects using ratarmount
    """
    filenames: list[str]
    filter_fn: Callable[[str], bool]

    _extracted_files: list[str]
    _mounted_dirs: list[Path]

    _supported_suffixes: ClassVar[list[str]] = None

    @classmethod
    def supported_suffixes(cls):
        """
        Get the list of suffixes for archives and compressed files which are supported
        """
        if not DAMAST_ARCHIVE_SUPPORT_AVAILABLE:
            return []

        if cls._supported_suffixes is not None:
            return cls._supported_suffixes

        cls._supported_suffixes = []
        for k, v in COMPRESSION_FORMATS.items():
            cls._supported_suffixes += v.extensions
        for k, v in ARCHIVE_FORMATS.items():
            cls._supported_suffixes += v.extensions
        return cls._supported_suffixes

    def __enter__(self) -> list[str]:
        """
        Enter function for the contextmanager
        """
        extracted_files = self.mount()
        if extracted_files:
            return extracted_files
        return self.filenames

    def __exit__(self):
        """
        Exit function for the contextmanager
        """
        self.umount()

    def __init__(self,
            filenames: list[str],
            filter_fn: Callable[[str], bool] | None = None):
        self.filenames = sorted(filenames)

        if filter_fn is None:
            self.filter_fn = lambda x: False
        else:
            self.filter_fn = filter_fn

        self._mounted_dirs = []
        self._extracted_files = []

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.umount()

    def ratarmount(self, file, target):
        """
        Call ratarmount to mount and archive
        """
        if not DAMAST_ARCHIVE_SUPPORT_AVAILABLE:
            raise RuntimeError("damast.utils.io.Archive: "
                    "cannot load archive."
                    " 'ratarmount' support is not available."
                )

        response = subprocess.run(["ratarmount","--recursive",file, target])
        if response.returncode != 0:
            raise RuntimeError(f"Mounting archive failed with exitcode {response.returncode}")

        self._mounted_dirs.append(target)


    def umount(self):
        """
        Umount the archive
        """
        for mounted_dir in list(reversed(self._mounted_dirs)):
            for count in range(0,5):
                time.sleep(0.5)
                response = subprocess.run(["fusermount", "-u", mounted_dir])
                if response.returncode == 0:
                    break
                else:
                    logger.debug(f"Retrying to unmount {mounted_dir}")

        for mounted_dir in self._mounted_dirs:
            if Path(mounted_dir).exists():
                shutil.rmtree(mounted_dir)

    def mount(self) -> list[str]:
        """
        Mount the archive (and potentially) inner archives to make the files accessible
        """
        if self._mounted_dirs:
            raise RuntimeError("Archive.mount: looks like these files are already mounted. Call 'umount' first")

        extracted_files = []
        local_mount = tempfile.mkdtemp(prefix=DAMAST_MOUNT_PREFIX)

        for file in self.filenames:
            if Path(file).suffix[1:] in Archive.supported_suffixes():
                logger.info(f"Archive.mount: found archive: {file}")
                target_mount = Path(local_mount) / Path(file).name
                target_mount.mkdir(parents=True, exist_ok=True)

                self.ratarmount(file, target_mount)

                decompressed_files = [x for x in Path(target_mount).glob("**/*") if Path(x).is_file()]
                for idx, x in enumerate(decompressed_files):
                    if self.filter_fn(x):
                        logger.debug(f"Archive.mount: ignoring unsupported file={x}")
                    else:
                        extracted_files += [ x ]
            elif self.filter_fn(file):
                logger.debug(f"Archive.mount: ignoring unsupported {file=}")
            else:
                # no extraction needed, and file suffix is supported
                extracted_files += [ file ]

        self._extracted_files = extracted_files
        return self._extracted_files

    def files(self):
        if not self._mounted_dirs and not self._extracted_files:
            raise RuntimeError("Archive.files: looks like nothing has been mounted. Call 'mount' first.")

        return self._extracted_files



DAMAST_MAJOR_VERSION = 0
DAMAST_MINOR_VERSION = 1
DAMAST_PATCH_LEVEL = 10

__version_info__ = (str(DAMAST_MAJOR_VERSION), str(DAMAST_MINOR_VERSION), str(DAMAST_PATCH_LEVEL))
__version__ = '.'.join(__version_info__)



"""
Module containing the functionality to explore and visualise data
"""
from pathlib import Path
from typing import List, Optional

from matplotlib import pyplot as plt

from damast.core.types import DataFrame

__all__ = ["plot_histograms",
           "plot_lat_lon",
           "PLOT_DPI"
           ]

#: Default DPI when plotting figures
PLOT_DPI: int = 300


def plot_lat_lon(*,
                 df: DataFrame,
                 output_dir: Path,
                 latitude_name: str = "LAT",
                 longitude_name: str = "LON",
                 filename_prefix: str = "lat-lon-scatter",
                 dpi: int = PLOT_DPI) -> Path:
    """
    Save a scatter plot of latitude longitude
    :param df:
    :param output_dir:
    :param longitude_name: Name of the longitude column in the dataframe
    :param latitude_name: Name of the latitude column in the dataframe
    :param filename_prefix: filename prefix for the output scatter plot
    :param dpi: DPI to save the final image with

    :return: Path to the file
    """
    if not isinstance(df, DataFrame):
        df = df.lazy()

    longitude = df.select(longitude_name).collect()
    latitude = df.select(latitude_name).collect()

    # .copy() to address ValueError: assignment destination is read-only
    # raised in numpy/ma/core.py:3528: in mask
    #      self.__setmask__(value)
    plt.scatter(x=longitude.to_numpy(),
                y=latitude.to_numpy(),
                alpha=1)
    plt.xlim(-180, 180)
    plt.ylim(-90, 90)
    plt.xlabel("longitude")
    plt.ylabel("latitude")
    filename = output_dir / f"{filename_prefix}.png"
    plt.savefig(filename, dpi=dpi)
    plt.close()

    return filename


def plot_histograms(*,
                    df: DataFrame,
                    output_dir: Path,
                    filename_prefix: str,
                    columns: Optional[List[str]] = None,
                    dpi: int = PLOT_DPI) -> Path:
    """
    Plot histograms for the current data frame in to directory defined through get_plot_dir()

    :param df: dataframe to generate the histograms for
    :param output_dir: output directory
    :param filename_prefix: filename prefix for the histograms
    :param dpi: plot the figure with this DPI (Dots Per Inch) setting
    :param columns: columns of the dataframe to use
    :return: Path to the plots' directory
    """
    if not isinstance(df, DataFrame):
        df = df.lazy()

    if columns is None:
        columns = df.columns

    # For each column plot the histogram
    for col_name in columns:
        if col_name not in df.columns:
            raise KeyError(f"plot_histogram: {col_name} is not an existing column,"
                           f" available are {','.join(df.columns)}")
        # Vaex uses the prefix "__" for hidden columns, i.e. columns that have either been removed, or
        # is a dependent of another column (say after converting radians to degrees)
        if col_name.startswith("__"):
            continue
        fig = plt.figure()

        df.select(col_name).collect().plot.hist()

        plt.tight_layout()
        path = output_dir / f"{filename_prefix}{col_name}.png"
        fig.savefig(path, dpi=dpi)
        plt.close(fig)
    return output_dir



DAMAST_HDF5_ROOT: str = "/dataframe"
DAMAST_HDF5_COLUMNS: str = "/dataframe/columns"

DAMAST_SPEC_SUFFIX: str = ".spec.yaml"
DAMAST_SUPPORTED_FILE_FORMATS: dict[str, list[str]] = {
    "parquet": [".parquet", ".pq"],
    "netcdf": [".nc", ".netcdf", ".nc4", ".netcdf4"],
    "hdf": [".h5", ".hdf", ".hdf5"],
    "csv": [".csv"],
}

DAMAST_MOUNT_PREFIX: str = "damast-mount"



import datetime
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from logging import INFO, Logger, getLogger
from random import choice, randint, random
from typing import Any, List

import pandas as pd
import polars as pl
from pyais.ais_types import AISType

from damast.core.metadata import DataCategory
from damast.core.types import DataFrame, XDataFrame
from damast.domains.maritime.ais import vessel_types
from damast.domains.maritime.ais.navigational_status import AISNavigationalStatus
from damast.domains.maritime.data_specification import (
    MMSI,
    ColumnName,
    CourseOverGround,
    SpeedOverGround,
    )

_log: Logger = getLogger(__name__)
_log.setLevel(INFO)

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
AIS_DATA_COLUMNS = [
    "mmsi",
    "lon",
    "lat",
    "date_time_utc",
    "sog",
    "cog",
    "true_heading",
    "nav_status",
    "rot",
    "message_nr",
    "source"
]

TRAJECTORY_MIN_SIZE = 10
TRAJECTORY_MAX_SIZE = 200

AISTestDataSpec = {
    "annotations": {"comment": "This is a autogenerated test data set"},
    "columns": [
        {"name": "mmsi", "is_optional": False, "category": DataCategory.STATIC},
        # , "value_range": CyclicMinMax(-180.0, 180.0)},
        {"name": "lon", "is_optional": False, "unit": "deg", "category": DataCategory.DYNAMIC},
        # , "value_range": CyclicMinMax(-90.0, 90.0)},
        {"name": "lat", "is_optional": False, "unit": "deg", "category": DataCategory.DYNAMIC},
        {"name": "date_time_utc", "is_optional": False, "category": DataCategory.DYNAMIC,
         "representation_type": "str"},
        {"name": "sog", "is_optional": False, "category": DataCategory.DYNAMIC},
        {"name": "cog", "is_optional": False, "category": DataCategory.DYNAMIC},
        {"name": "true_heading", "is_optional": False, "category": DataCategory.DYNAMIC},
        {"name": "nav_status", "is_optional": False, "category": DataCategory.DYNAMIC},
        {"name": "rot", "is_optional": False, "category": DataCategory.DYNAMIC},
        {"name": "message_nr", "is_optional": False, "category": DataCategory.DYNAMIC},
        {"name": "source", "is_optional": False, "category": DataCategory.DYNAMIC},
    ]
}


class AISTestData:
    dataframe: DataFrame = None

    def __init__(self,
                 number_of_trajectories: int,
                 min_length: int = TRAJECTORY_MIN_SIZE,
                 max_length: int = TRAJECTORY_MAX_SIZE):

        self.number_of_trajectories = number_of_trajectories
        self.min_length = min_length
        self.max_length = max_length

        self.dataframe: DataFrame = self._generate_data()

    @staticmethod
    def generate_trajectory(min_size: int, max_size: int) -> List[List[Any]]:
        """Generate a trajectory for a single vessel

        :param min_size: Minimum size of trajectory
        :param max_size: Maximum size of trajectory
        :return: A nested list describing the trajectory
        """
        # Create some invalid messages as well
        range_ext = 0.05
        mmsi = randint(int((1-range_ext)*MMSI.min_value), int((1+range_ext)*MMSI.max_value))
        lat_start = (random() * 180.0) - 90.0
        lon_start = (random() * 360.0) - 180.0

        last_day = datetime.datetime(year=2022, month=12, day=31)
        start_time = randint(0, int(last_day.timestamp()))

        sog_start = random() * SpeedOverGround.min_value / SpeedOverGround.max_value
        cog_start = random() * CourseOverGround.min_value / CourseOverGround.max_value
        heading_start = cog_start

        rot_start = 0.0

        # Choose a random message ID from a subset of available types
        msg_id = choice([AISType.POS_CLASS_A1, AISType.POS_CLASS_A2, AISType.POS_CLASS_A3])

        trajectory_length = randint(min_size, max_size)
        trajectory = []
        for _ in range(0, trajectory_length):
            lat_start += random() * 0.1 - 0.05
            lon_start += random() * 0.1 - 0.05

            # message within 2 min
            start_time += random() * 120.0

            sog_start += random() * 4 - 2
            cog_start += random() * 1 - 0.5
            heading_start = cog_start + random() / 10.0

            nav_status = choice([
                AISNavigationalStatus.EngagedInFishing.value,
                AISNavigationalStatus.UnderWayUsingEngine.value,
                AISNavigationalStatus.AtAnchor.value
            ])
            # 5 percent of all messages should be missing a time-stamp
            q = random()
            if q > 0.1:
                timestamp = datetime.datetime.fromtimestamp(start_time).strftime(
                    DATETIME_FORMAT)    # 2022-09-01 12:55:35
            else:
                timestamp = None
            trajectory.append(
                [mmsi,
                 lon_start,
                 lat_start,
                 timestamp,
                 sog_start,
                 cog_start,
                 heading_start,
                 nav_status,
                 rot_start,
                 msg_id,
                 choice(["s", "g"])  # Create some ground data that should be removed in processing
                 ])
        return trajectory

    def _generate_data(self) ->DataFrame:
        df = None
        for i in range(0, self.number_of_trajectories):
            trajectory = AISTestData.generate_trajectory(min_size=self.min_length, max_size=self.max_length)
            t_df = pd.DataFrame(trajectory, columns=AIS_DATA_COLUMNS)
            if df is None:
                df = t_df
            else:
                df = pd.concat([df, t_df], axis=0, ignore_index=True)
        # shuffle rows
        df = df.sample(frac=1, axis="index").reset_index(drop=True)
        return XDataFrame.from_pandas(df)

    def generate_vessel_type_data(self) -> DataFrame:
        """Generate a :class:`DataFrame` with data imitating vessel-type info

        Returns:
            A dataframe with an `ColumnName.MMSI` column and a `ColumnName.VESSEL_TYPE` column.
        """
        vessel_type_data = []
        for mmsi in self.dataframe[ColumnName.MMSI.lower()].unique():
            vessel_type_data.append([mmsi, vessel_types.Fishing.to_id()])
        df = pd.DataFrame(vessel_type_data, columns=[ColumnName.MMSI, ColumnName.VESSEL_TYPE])
        return XDataFrame.from_pandas(df)

    def generate_fishing_vessel_type_data(self) -> DataFrame:
        """Generate a `vaex.Dataframe` with data imitating vessel info from Global Fishing Watch

        Returns:
            A dataframe with an `ColumnName.MMSI` (lower-cased) column and a `ColumnName.VESSEL_TYPE` column.
        """
        fishing_vessel_type_data = []
        for mmsi in self.dataframe[ColumnName.MMSI.lower()].unique():
            fishing_vessel_type_data.append([mmsi, choice([
                vessel_types.DriftingLonglines.to_id(),
                vessel_types.PoleAndLine.to_id(),
                vessel_types.PotsAndTraps.to_id()
            ])])

        df = pd.DataFrame(fishing_vessel_type_data, columns=[ColumnName.MMSI.lower(), ColumnName.VESSEL_TYPE_GFW])

        return XDataFrame.from_pandas(df)

    def generate_anchorage_type_data(self) -> DataFrame:
        # Generate anchorages close to starting points of vessel trajectories
        mmsi = ColumnName.MMSI.lower()
        first = self.dataframe.group_by(mmsi).first()
        first = first.rename({ "lat": "latitude", "lon": "longitude"})
        df = first.with_columns(
                pl.col("latitude")*1.05,
                pl.col("longitude")*-0.95,
            )
        return df


if __name__ == "__main__":
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            description="A simple generator for ais data that can be used as input to "
                                        + "data processing stage")

    parser.add_argument("-t", "--number-of-trajectories", default=100, type=int)
    parser.add_argument("-o", "--output", default="test-data.hdf5", type=str)

    args = parser.parse_args()

    ais_test_data = AISTestData(number_of_trajectories=args.number_of_trajectories)
    ais_test_data.dataframe.export_hdf5(args.output)
    _log.info("Written: {args.output}")



import logging
import shutil
import subprocess
import tempfile
from argparse import ArgumentParser
from pathlib import Path

from tqdm import tqdm

from damast.cli.base import BaseParser
from damast.core.dataframe import AnnotatedDataFrame
from damast.core.metadata import MetaData, ValidationMode
from damast.utils.io import Archive

logger = logging.getLogger(__name__)

class DataConvertParser(BaseParser):
    """
    Argparser for converting CSV files to AnnotatedDataframes (.parquet and .yml files)

    :param parser: The base parser
    """

    def __init__(self, parser: ArgumentParser):
        super().__init__(parser=parser)

        parser.description = "damast convert - data conversion subcommand called"
        parser.add_argument("-f", "--files",
                            help="Files or patterns of the (annotated) data file that should be converted",
                            nargs="+",
                            type=str,
                            required=True
                            )
        parser.add_argument("-m", "--metadata-input",
                            help="The metadata input file",
                            default=None,
                            required=False
                            )
        parser.add_argument("-o", "--output-file",
                            help="The output file either: .parquet, .hdf5",
                            required=False
                            )
        parser.add_argument("--output-dir",
                            help="The output directory",
                            required=False,
                            )
        parser.add_argument("--output-type",
                            help="The output file type: .parquet (default) or .hdf5 ",
                            default=".parquet",
                            required=False,
                            )
        parser.add_argument("--validation-mode",
                            default="readonly",
                            choices=[x.value.lower() for x in ValidationMode],
                            help="Define the validation mode")
    def execute(self, args):
        super().execute(args)

        files_stats = self.get_files_stats(args.files)
        print(f"Loading dataframe ({files_stats.number_of_files} files) of total size: {files_stats.total_size} MB")

        if args.output_dir and args.output_file:
            raise ValueError("--output-dir and --output-file cannot be used together")

        with Archive(filenames=args.files) as input_files:
            files = [x for x in input_files if AnnotatedDataFrame.get_supported_format(Path(x).suffix)]
            if not files:
                raise RuntimeError(f"Conversion is not supported for input files: {input_files=}")

            created_files = []
            if args.output_dir:
                # Create multiple output files
                output_dir = Path(args.output_dir)
                output_dir.mkdir(parents=True, exist_ok=True)

            for file in tqdm(files, desc="Files"):
                adf = AnnotatedDataFrame.from_file(
                        filename=file,
                        metadata_required=False,
                    )

                if args.output_dir:
                    output_file = output_dir / f"{Path(file).stem}{args.output_type}"
                elif args.output_file:
                    output_file = Path(args.output_file)
                else:
                    # Use current directory as output dir
                    output_file = Path(Path(file).with_suffix(args.output_type).name)

                adf.save(filename=output_file)
                created_files.append(output_file)

                if args.metadata_input:
                    if not Path(args.metadata_input).exists():
                        raise FileNotFoundError(f"metadata-input: '{args.metadata_input}' does not exist")
                    metadata = MetaData.load_yaml(filename=args.metadata_input)

                    try:
                        validation_mode = ValidationMode[args.validation_mode.upper()]
                    except KeyError:
                        raise ValueError(f"--validation-mode has invalid argument."
                                         f" Select from: {[x.value.lower() for x in ValidationMode]}")
                    adf._metadata = metadata
                    adf.validate_metadata(validation_mode)

                print(f"Filename: {output_file.resolve()}")
                print(adf.head(10).collect())

            print(f"Written: {created_files}")




from argparse import ArgumentParser
from pathlib import Path

from damast.cli.base import BaseParser
from damast.core.dataframe import AnnotatedDataFrame
from damast.core.dataprocessing import DAMAST_PIPELINE_SUFFIX, DataProcessingPipeline
from damast.utils.io import Archive


class DataProcessingParser(BaseParser):
    def __init__(self, parser: ArgumentParser):
        super().__init__(parser=parser)

        parser.description = "damast process - apply an existing pipeline"

        parser.add_argument("--input-data",
                            help="Input file(s) to process",
                            nargs="+",
                            type=str,
                            required=True
        )
        parser.add_argument("--pipeline", help="Pipeline (*.damast.ppl) file to apply to the data", required=True)

        parser.add_argument("--output-file",
                        help="Save the result in the given (*.parquet) file",
                        required=False)

    def execute(self, args):
        super().execute(args)

        with Archive(filenames=args.input_data) as input_files:
            files = [x for x in input_files if AnnotatedDataFrame.get_supported_format(Path(x).suffix)]
            if not files:
                raise RuntimeError(f"Processing is not possible for input files: {input_files=}")

            adf = AnnotatedDataFrame.from_files(
                    files,
                    metadata_required=False
            )

            pipeline_path = Path(args.pipeline)
            if not pipeline_path.exists():
                raise FileNotFoundError(f"Pipeline {pipeline_path} does not exist")

            if not str(pipeline_path).endswith(DAMAST_PIPELINE_SUFFIX):
                raise ValueError(f"File suffix of pipeline file is not matching {DAMAST_PIPELINE_SUFFIX}")


            pipeline = DataProcessingPipeline.load(pipeline_path)
            new_adf = pipeline.transform(adf)

            print(new_adf.head().collect())
            print(new_adf.tail().collect())

            if args.output_file:
                path = Path(args.output_file)
                path.parent.resolve().mkdir(parents=True, exist_ok=True)

                new_adf.save(filename=path)
                print(f"Saved {path.resolve()}")



"""Main module"""

from damast import data_handling  # , domains
from damast import cli, core
from damast.core import artifacts, describe, input, output

from .version import __version__, __version_info__

__all__ = [
    "artifacts",
    "cli",
    "core",
    "data_handling",
    "describe",
#    "domains",
    "input",
    "output",
    "__version__",
    "__version_info__"
]



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import ClassVar

import numpy as np
import polars
import polars.api
from polars import LazyFrame
from pydantic import ValidationError

from .data_description import NumericValueStats

logger = logging.getLogger(__name__)

VAEX_HDF5_ROOT: str = "/table"
VAEX_HDF5_COLUMNS: str = f"{VAEX_HDF5_ROOT}/columns"

class Meta(type):
    _base_impl: ClassVar["str"] = "polars"

    def __getattr__(cls, attr_name):
        if cls._base_impl == 'polars':
            return getattr(polars, attr_name)

        raise AttributeError(f"'{cls.__name__}' has not attribute '{attr_name}'")

@polars.api.register_dataframe_namespace("compat")
@polars.api.register_lazyframe_namespace("compat")
class PolarsDataFrame(metaclass=Meta):
    _dataframe: LazyFrame
    _polars_dataframe: PolarsDataFrame
    _dataframe_collected: polars.DataFrame

    def __init__(self, df: LazyFrame | polars.DataFrame):
        if type(df) == polars.DataFrame:
            self._dataframe = df.lazy()
        else:
            self._dataframe = df

        self._dataframe_collected = None
        self._polars_dataframe = None

    @property
    def dataframe(self) -> PolarsDataFrame:
        """
        Allows to access the underlying dataframe directly.

        .. note::
            AnnotatedDataFrame behaves like a ``polars.LazyFrame``, so typically you will not need to access the
            dataframe through this property.

        :return: The underlying dataframe
        """
        if self._polars_dataframe is None or self._dataframe is not self._polars_dataframe._dataframe:
            self._polars_dataframe = PolarsDataFrame(self._dataframe)

        return self._polars_dataframe

    def collected(self):
        if self._dataframe_collected is None:
            self._dataframe_collected = self._dataframe.collect()

        return self._dataframe_collected

    def is_string(self, column_name: str) -> bool:
        return str(self.dtype(column_name)).lower().startswith("str")

    def is_numeric(self, column_name: str) -> bool:
        return self.dtype(column_name).is_numeric()

    def __getitem__(self, column_name: str):
        """
        Make dataframe subscriptable and behave more like the :class:`pandas.DataFrame`.

        :param item: Name of the key when using [] operators
        :return: item/column from the underlying vaex.dataframe
        """
        return self._dataframe.select(column_name)

    @property
    def column_names(self) -> list[str]:
        """
        Get all column names (without collecting the full dataframe)
        """
        return self._dataframe.collect_schema().names()

    def dtype(self, column_name: str) -> polars.datatypes.DataType:
        """
        Get column dtype (without collecting the full dataframe)
        """
        idx = self.column_names.index(column_name)
        return self._dataframe.collect_schema().dtypes()[idx]

    def set_dtype(self, column_name, representation_type) -> PolarsDataFrame:
        """
        Set the dtype for a column to the given representation type.
        Using polars cast functionality
        :return: The updated object
        """
        if representation_type == np.int64:
            representation_type = polars.Int64

        self._df = self._dataframe.with_columns(polars.col(column_name).cast(representation_type).alias(column_name))
        return self

    def minmax(self, column_name: str) -> Tuple[any, any]:
        """
        Tuple of min and max values of the given column
        """
        try:
            result = self._dataframe.select([
                    polars.col(column_name).min().alias("min_value"),
                    polars.col(column_name).max().alias("max_value")
                ]).collect()
        except polars.exceptions.InvalidOperationError as e:
            raise ValueError(f"damast.core.polars_dataframe.minmax: cannot compute min/max for {column_name}") from e

        min_value = result["min_value"][0]
        max_value = result["max_value"][0]

        return min_value, max_value

    def categories(self, column_name: str, max_count: int = 100) -> list[str]:
        categories = self._dataframe.select(column_name).unique().sort(by=column_name).collect()[:,0].to_list()
        if len(categories) <= max_count:
            timepoint_like = 0
            for c in categories[:10]:
                if c and re.search(r"[0-9]{2}:[0-9]{2}", c) is not None:
                    timepoint_like += 1

            if timepoint_like < 3:
                return categories

        return None


    def minmax_stats(self, column_names: list[str]) -> dict[str, dict[str, any]]:
        """
        Tuple of min and max values of the given column
        """
        fields = []
        for column in column_names:
            fields.extend([
                polars.col(column).min().alias(f"{column}_min_value"),
                polars.col(column).max().alias(f"{column}_max_value"),
                polars.col(column).mean().alias(f"{column}_mean"),
                polars.col(column).std().alias(f"{column}_stddev"),
                polars.col(column).count().alias(f"{column}_total_count"),
                polars.col(column).null_count().alias(f"{column}_null_count")
            ])

        result = self._dataframe.select(
                fields
        ).collect()

        results = {}
        for column in column_names:
            min_value = result[f"{column}_min_value"][0]
            max_value = result[f"{column}_max_value"][0]
            try:
                stats = NumericValueStats(
                    mean=result[f"{column}_mean"][0],
                    stddev=result[f"{column}_stddev"][0],
                    total_count=result[f"{column}_total_count"][0],
                    null_count=result[f"{column}_null_count"][0],
                )
            except ValidationError as e:
                logger.warning(f"Unable to compute stats for {column=}")
                logger.debug(f"Validation error for {column=} -- {e}")
                stats = None

            results[column] = {
                    "min_value": min_value,
                    "max_value": max_value,
                    "stats": stats
            }

        return results


    def stats(self, column_name: str) -> NumericValueStats:
        result = self._dataframe.select([
            polars.col(column_name).mean().alias("mean"),
            polars.col(column_name).std().alias("stddev"),
            polars.col(column_name).count().alias("total_count"),
            polars.col(column_name).null_count().alias("null_count")
        ]).collect()

        return NumericValueStats(
                mean=result['mean'][0],
                stddev=result['stddev'][0],
                total_count=result['total_count'][0],
                null_count=result['null_count'][0]
        )

    def __getattr__(self, attr_name):
        """
        Ensure that this object behaves like a :class:`polars.LazyFrame`.

        :param attr_name: Attribute / Name of column
        :return: The column data
        """
        # allow dataframe.col_one
        if attr_name in self.column_names:
            return self._dataframe.select(attr_name)

        if attr_name in ["__setstate__", "__getstate__"]:
            raise AttributeError(f"{self.__class__.__name__}.__getattr__: {attr_name} does not exist")

        """ Called for failed attribute accesses so forwarding to underlying polars frame """
        return getattr(self._dataframe, attr_name)

    def __setitem__(self, key, values):
        """
        Set the column for the annotated dataframe, and allow to behave like the polars.Dataframe.

        :param key: Column name
        :param value: Value to set the column to
        """
        if type(values) == polars.LazyFrame:
            values = values.collect().to_numpy()

        self._dataframe = self._dataframe.with_columns(
                    polars.Series(
                        name=key,
                        values=values
                    )
                )

    def __len__(self) -> int:
        """
        Get the length of the (underlying) dataframe.

        :return: Length of the dataframe
        """
        return len(self.collected())

    def equals(self, other: PolarsDataFrame) -> bool:
        return self.collected().equals(other.collected())

    def open(path: str | Path, sep = ',') -> DataFrame:
        path = Path(path)
        if path.suffix == ".csv":
            return polars.scan_csv(path, sep=sep, quotechar=None, infer_schema_length=None)
        elif path.suffix in [".h5", ".hdf5"]:
            import pandas as pd

            from damast.core.metadata import DAMAST_HDF5_ROOT

            pandas_df = pd.read_hdf(path, key=DAMAST_HDF5_ROOT)
            return polars.from_pandas(pandas_df)
        elif path.suffix in [".pq", ".parquet"]:
            return polars.scan_parquet(path)

        raise ValueError(f"{cls.__name__}.load_data: Unsupported input file format {path.suffix}")

    @classmethod
    def from_vaex_hdf5(cls, path: str | Path) -> Tuple[DataFrame, MetaData]:
        """
        Load hdf5 file and (damast) metadata if found in the file.
        """
        # avoid circular dependencies
        from damast.core.annotations import Annotation
        from damast.core.metadata import DataSpecification, MetaData
        try:
            import tables
        except ImportError as e:
            raise RuntimeError("Could not load pytables -- "
                    "please install 'tables' to use hdf5 functionality") from e

        annotations = []
        column_specifications = []
        with tables.open_file(str(path)) as hdf5file:
            if not VAEX_HDF5_ROOT in hdf5file:
                raise TypeError(f"This HDF5 file '{hdf5file}' has not been exported with vaex")

            table_attrs = hdf5file.get_node(VAEX_HDF5_ROOT)._v_attrs
            for key in table_attrs._f_list():
                value = table_attrs[key]
                annotations.append(
                        Annotation(name=key, value=value)
                )

            data = {}
            for column in hdf5file.get_node(VAEX_HDF5_COLUMNS):
                raw_data = column.data.read()
                if isinstance(raw_data, np.ndarray) and np.issubdtype(raw_data.dtype, np.bytes_):
                    raw_data = [x.decode('utf-8') for x in raw_data]
                data[column._v_name] = raw_data

                data_specification_dict = {}
                for key in column._v_attrs._f_list():
                    data_specification_dict[key] = column._v_attrs[key]

                if data_specification_dict:
                    column_spec = DataSpecification.from_dict(data=data_specification_dict)
                    column_specifications.append(column_spec)

        metadata = None
        if column_specifications or annotations:
            metadata = MetaData(columns=column_specifications, annotations=annotations)

        return polars.LazyFrame(data), metadata

    @classmethod
    def import_netcdf(cls, path: list[str|Path]) -> Tuple[DataFrame, dict[str, MetaData]]:
        try:
            import dask
            import xarray
        except ImportError as e:
            raise RuntimeError("Loading netcdf files requires to 'pip install xarray dask'"
                    ", additionally either netcdf4 or h5netcdf") from e

        try:
            import pandas as pd
        except ImportError as e:
            raise RuntimeError("Loading netcdf files requires pandas, "
                " e.g., run 'pip install pandas'") from e

        dataframes = []
        for f in path:
            ds = xarray.open_dataset(f)
            dataframes.append( ds.to_dataframe().reset_index() )
        pandas_df = pd.concat(dataframes, ignore_index=True).reset_index()
        df = polars.from_pandas(pandas_df)

        return df.lazy(), {}

    @classmethod
    def import_hdf5(cls, files: str | Path | list[str|Path]) -> Tuple[DataFrame, dict[str,MetaData]]:
        """
        Import a dataframe stored as HDF5.

        This method tries to load using pandas first, then falls back to reading a vaex-based format
        using pytables.
        """
        try:
            import tables
        except ImportError as e:
            raise RuntimeError("Could not load pytables -- please install 'tables' to use hdf5 functionality") from e

        try:
            import pandas
        except ImportError as e:
            raise RuntimeError("Could not load pandas -- please install 'pandas' to use hdf5 functionality") from e

        if type(files) != list:
            files = [files]

        try:
            import warnings

            # Avoid output pollution and ignore the following message
            #     tables/attributeset.py:295: DataTypeWarning: Unsupported type
            #     for attribute 'is_optional' in node 'height'. Offending HDF5
            #     class: 8
            warnings.filterwarnings("ignore", message="Unsupported type for attribute .*")

            data_frames = []
            for filename in files:
                data_frames.append( pandas.read_hdf(str(filename)) )
            pandas_df = pandas.concat(data_frames, ignore_index=True)

            df = polars.from_pandas(pandas_df)
            warnings.resetwarnings()

            return df.lazy(), {}
        except tables.exceptions.NoSuchNodeError as e:
            logger.debug(f"HDF5 {filename} cannot be imported with pandas")

        if len(files) > 1:
            raise RuntimeError("Loading from vaex is only supported with one file at a time")

        path = Path(files[0])
        df, metadata = cls.from_vaex_hdf5(path)

        return df, { path: metadata }

    @classmethod
    def export_hdf5(cls, df: DataFrame, path: str | Path) -> Path:
        """
        Export the dataframe as hdf5. Please use only if really needed, otherwise, stick with the
        default format (parquet).
        """
        try:
            import pandas
        except ImportError as e:
            raise RuntimeError("Could not load pandas -- please install pandas to use hdf5 functionality")

        from damast.core.metadata import DAMAST_HDF5_ROOT

        if isinstance(df, polars.dataframe.DataFrame):
            df = df.lazy()

        pandas_df = df.collect().to_pandas()
        pandas_df.to_hdf(path, key=DAMAST_HDF5_ROOT)
        return path






from .augmenters import AddMissingAISStatus, AddVesselType, ComputeClosestAnchorage
from .features import DeltaDistance, Speed

__all__ = [
    "AddMissingAISStatus",
    "DeltaDistance",
    "AddVesselType",
    "ComputeClosestAnchorage",
    "Speed"
]



import numpy as np
import polars as pl

import damast.core
from damast.core import AnnotatedDataFrame, DataSpecification
from damast.core.dataprocessing import PipelineElement
from damast.core.types import DataFrame
from damast.domains.maritime.math import great_circle_distance

__all__ = [
    "DeltaDistance",
    "Speed"
]


class DeltaDistance(PipelineElement):
    """
    Given a dataframe with `(latitude, longitude)` data, group messages by given column,
    and sort them by another column. Then compute the distance between two messages, using
    the :func:`damast.domains.maritime.math.great_circle_distance`.

    :param x_shift: True if one should compute the difference in latitude
    :param y_shift: True if one should compute the difference in longitude

    .. note::
        If both :code:`x_shift` and :code:`y_shift` is :code:`True`, one computes the distance between two coordinates.
    """
    _x_shift: bool
    _y_shift: bool

    def __init__(self, x_shift: bool, y_shift: bool):
        self._x_shift = x_shift
        self._y_shift = y_shift

    @property
    def x_shift(self):
        return self._x_shift

    @property
    def y_shift(self):
        return self._y_shift

    @damast.core.describe("Compute the distance between lat/lon given positions")
    @damast.core.input({"group": {"representation_type": int},
                        "sort": {},
                        "x": {"unit": "deg" },
                        "y": {"unit": "deg" }})
    @damast.core.output({"out": {"description": "distance to previous coordinate", "unit": "km"}})
    def transform(self,
                  df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Compute distance between adjacent messages
        """
        dataframe = df._dataframe

        in_x = self.get_name("x")
        in_y = self.get_name("y")
        shift_x = f"{in_x}_shifted"
        shift_y = f"{in_y}_shifted"

        tmp_column = f"{self.__class__.__name__}_tmp"
        assert tmp_column != self.get_name("out")
        if tmp_column in dataframe.compat.column_names:
            raise RuntimeError(f"{self.__class__.__name__}.transform: Dataframe contains {tmp_column}")

        # https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.shift.html
        # shift
        # Group data and sort it. Only shift data within the same group
        groups = dataframe.sort(self.get_name('sort'))\
                    .group_by(self.get_name("group"), maintain_order=True)\
                    .agg(\
                        pl.col("*"),
                        (pl.col(in_x).shift(1)).alias(shift_x),\
                        (pl.col(in_y).shift(1)).alias(shift_y)\
                    )

        expanded = groups.explode(pl.exclude(self.get_name('group')))\
                      .drop_nulls([shift_x, shift_y])\
                      .with_columns(
                        pl.struct(
                            in_x, in_y,
                            shift_x, shift_y).map_elements(
                                lambda x : great_circle_distance(x[in_x], x[in_y], x[shift_x], x[shift_y]),
                                return_dtype=float
                            ).alias(self.get_name('out'))
                      )

        # Drop/Hide unused columns
        df._dataframe = expanded.drop([shift_x, shift_y])
        return df


class Speed(PipelineElement):
    """
    Given a dataframe with delta time and delta distance compute the speed
    """
    @damast.core.describe("Compute the speed")
    @damast.core.input({"delta_distance": {"unit": "km"},
                        "delta_time": {}})
    @damast.core.output({"speed": {"description": "speed of object", "unit": "km / h"}})
    def transform(self,
                  df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Compute distance between adjacent messages
        """
        dataframe = df._dataframe

        delta_distance = self.get_name("delta_distance")
        delta_time = self.get_name("delta_time")

        df._dataframe = dataframe.with_columns(
                (pl.col(delta_distance) / (pl.col(delta_time).dt.total_seconds()/3600)).alias("speed")
                        )

        return df



"""
This module contains data range definitions.
"""
import warnings

from .data_description import CyclicMinMax, DataElement, DataRange, ListOfValues, MinMax

__all__ = ["CyclicMinMax", "DataElement", "DataRange", "ListOfValues", "MinMax"]

warnings.warn("damast.core.datarange has been deprecated and will be remove in future versions. Use damast.core.data_description instead")




"""
Module containing maritime specific functionality, including transformers and mathematical operators.
"""
from damast.domains.maritime import ais, math, transformers

__all__ = ["ais", "math", "transformers"]



"""
Module which collects transformers that add / augment the existing data
"""
from pathlib import Path
from typing import Callable, List, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl

import damast.core
import damast.data_handling.transformers.augmenters as augmenters
from damast.core.dataprocessing import PipelineElement
from damast.core.types import DataFrame, XDataFrame
from damast.data_handling.transformers.augmenters import BallTreeAugmenter
from damast.domains.maritime.ais.navigational_status import AISNavigationalStatus
from damast.domains.maritime.ais.vessel_types import VesselType
from damast.domains.maritime.math.spatial import EARTH_RADIUS

__all__ = [
    "AddMissingAISStatus",
    "ComputeClosestAnchorage"
]


class ComputeClosestAnchorage(PipelineElement):
    """
    Compute the closest anchorage given a data-set with all closest anchorages

    :param dataset: Path to data-set with closest anchorages
    :param columns: Names of columns used to define the distance to anchorage (The data should be in degrees)
    :param sep: Separator used in dataset if dataset is a csv file
    """
    _function: Callable[[npt.NDArray[np.float64], npt.NDArray[np.float64]], npt.NDArray[np.float64]]

    dataset: str
    columns: list[str]
    sep: str

    @property
    def deg2rad(self) -> float:
        return np.pi/180.0

    def __init__(self,
                 dataset: str | Path,
                 columns: list[str],
                 sep: str = ";"):
        self.dataset = dataset
        self.columns = columns
        self.sep = sep

        if type(dataset) in [str, Path]:
            _dataset = self.load_data(dataset, sep)
        else:
            _dataset = dataset

        radian_dataset = [
                _dataset.select((pl.col(x)*self.deg2rad).alias(x))[:,0] for x in columns
        ]

        self._function = BallTreeAugmenter(np.vstack(radian_dataset).T, "haversine")

    @classmethod
    def load_data(cls,
                  filename: Union[str, Path], sep: str) -> DataFrame:
        """
        Load dataset from file

        :param filename: The input file (or path)
        :param sep: Separator in csv
        :return: A `pandas.DataFrame` where each row has a column MMSI and vessel_type
        """
        try:
            return XDataFrame.open(path=filename, sep=sep)
        except FileNotFoundError as e:
            raise RuntimeError(f"{cls}: Vessel type information not accessible. File {vessel_type_csv} not found")

    @damast.core.describe("Compute distance from dataset to closest anchorage")
    @damast.core.input({"x": {"representation_type": float, "unit": damast.core.units.units.deg},
                        "y": {"representation_type": float, "unit": damast.core.units.units.deg}})
    @damast.core.output({"distance": {"representation_type": float, "unit": damast.core.units.units.km}})
    def transform(self, df:  damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        dataframe = df._dataframe

        x_name = self.get_name('x')
        y_name = self.get_name('y')

        # Transform latitude and longitude to radians
        dataframe = dataframe.with_columns(
            (pl.col(x_name)*self.deg2rad).alias(f"{x_name}_rad")
        )
        dataframe = dataframe.with_columns(
            (pl.col(y_name)*self.deg2rad).alias(f"{y_name}_rad")
        )

        distance = self.get_name('distance')
        dataframe = dataframe.with_columns(
            pl.struct(f"{x_name}_rad", f"{y_name}_rad").map_batches(
                lambda x: self._function(x.struct.field(f"{x_name}_rad"), x.struct.field(f"{y_name}_rad")),
                return_dtype=float
            ).alias(distance)
        )

        # Multiply distance column by earth radius
        dataframe = dataframe.with_columns(
            (pl.col(distance)*EARTH_RADIUS).alias(distance)
        )
        #dataframe.units[self.get_name("distance")] = damast.core.units.units.km

        # Drop/hide conversion columns
        df._dataframe = dataframe.drop([f"{self.get_name('x')}_rad", f"{self.get_name('y')}_rad"])
        return df


class AddMissingAISStatus(augmenters.AddUndefinedValue):
    """For a given column replace rows with missing entries with
    :attr:`damast.domains.maritime.ais.AISNavigationalStatus.Undefined`.
    """

    def __init__(self):
        self._fill_value = int(AISNavigationalStatus.Undefined)

    @damast.core.describe("Fill missing AIS status")
    @damast.core.input({"x": {"representation_type": int}})
    @damast.core.output({"x": {"representation_type": int}})
    def transform(self, df: damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        return super().transform(df)


class AddVesselType(augmenters.JoinDataFrameByColumn):
    """
    Add the vessel type (as integer value) based on external data-set and corresponding to ::class::`VesselType`.

    The vessel type might for instance come from the [global fishing watch database](https://globalfishingwatch.org/).
    It must be either a string corresponding to the snake_case class name, or the corresponding integer

    :param right_on: Name in data-set column to use for merging datasets
    :param dataset_column: Column to add to input dataframe
    :param dataset: Dataset or path to dataset
    :raise ValueError: when the input column that shall extend the dataset is neither int nor str
    """

    def __init__(self,
                 right_on: str,
                 dataset_column: str,
                 dataset: Union[str, Path, DataFrame]
                 ):

        if type(dataset) in [str, Path]:
            dataset = XDataFrame.open(path=dataset)

        column_dtype = XDataFrame(dataset).dtype(dataset_column)
        name = f"{dataset_column}_mapped"
        if str in [column_dtype, column_dtype.to_python()]:
            # VesselTypes should be mapped to integers
            mapping = VesselType.get_mapping()
            dataset = dataset.with_columns(
                pl.col(dataset_column).replace_strict(mapping).alias(name)
            )
        elif int in [column_dtype, column_dtype.to_python()]:
            dataset = dataset.with_columns(
                    pl.col(dataset_column).alias(name)
            )
        else:
            raise ValueError(f"{self.__class__.__name__}.__init__: dtype of column '{dataset_column}',"
                             f" must be either int or str, but was '{column_dtype}'")
        super().__init__(dataset=dataset, right_on=right_on, dataset_column=name)

    @damast.core.describe("Add vessel-type from other dataset to current dataset, "
                          "where the input x is the linking identifier")
    @damast.core.input({"x": {}})
    @damast.core.output({"out": {"representation_type": int}})
    def transform(self, df: damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        return super().transform(df)



"""
Module containing decorators and classes to model data processing pipelines
"""
from __future__ import annotations

import copy
import functools
import importlib
import inspect
import re
import tempfile
from abc import abstractmethod
from datetime import datetime, timezone
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml

import damast.version
from damast.core.pipeline import Pipeline
from damast.core.transformations import Transformer

from .dataframe import AnnotatedDataFrame
from .formatting import DEFAULT_INDENT
from .metadata import ArtifactSpecification, DataSpecification, MetaData

__all__ = [
    "artifacts",
    "input",
    "output",
    "describe",
    "DataProcessingPipeline",
    "DECORATED_ARTIFACT_SPECS",
    "DECORATED_DESCRIPTION",
    "DECORATED_INPUT_SPECS",
    "DECORATED_OUTPUT_SPECS",
    "PipelineElement",
]

_log: Logger = getLogger(__name__)

DECORATED_DESCRIPTION = "_damast_description"
"""Attribute description for :func:`describe`"""

DECORATED_ARTIFACT_SPECS = "_damast_artifact_specs"
"""Attribute description for :func:`artifacts`"""

DECORATED_INPUT_SPECS = "_damast_input_specs"
"""Attribute description for :func:`input`"""

DECORATED_OUTPUT_SPECS = "_damast_output_specs"
"""Attribute description for :func:`output`"""

DAMAST_PIPELINE_SUFFIX: str = ".damast.ppl"
"""Suffix of :class:`DataProcessingPipeline` files created in :func:`DataProcessingPipeline.save`
and used by :func:`DataProcessingPipeline.load`"""

VAEX_STATE_SUFFIX: str = ".vaex-state.json"
"""Suffix of :class:`DataProcessingPipeline` files created in :func:`DataProcessingPipeline.save_state`
and used by :func:`DataProcessingPipeline.load_state`"""


def _get_dataframe(*args, **kwargs) -> AnnotatedDataFrame:
    """
    Extract the dataframe from positional or keyword arguments

    :param args: positional arguments
    :param kwargs: keyword arguments
    :return: The annotated data frame
    :raise KeyError: if a positional argument does not exist and keyword :code:`df` is missing
    :raise TypeError: if the kwargs 'df' is not an AnnotatedDataFrame
    """
    _df: AnnotatedDataFrame
    if len(args) >= 2 and isinstance(args[1], AnnotatedDataFrame):
        _df = args[1]
    elif "df" not in kwargs:
        raise KeyError("Missing keyword argument 'df' to define the AnnotatedDataFrame")
    elif not isinstance(kwargs["df"], AnnotatedDataFrame):
        raise TypeError("Argument 'df' is not an AnnotatedDataFrame")
    else:
        _df = kwargs["df"]
    return _df


def describe(description: str):
    """
    Specify the description for the transformation for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param description: description of the action
    """

    def decorator(func):
        setattr(func, DECORATED_DESCRIPTION, description)
        return func

    return decorator


def input(requirements: Dict[str, Any]):
    """
    Specify the input for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param requirements: List of input requirements
    """

    required_input_specs = DataSpecification.from_requirements(
        requirements=requirements
    )

    def decorator(func):
        setattr(func, DECORATED_INPUT_SPECS, required_input_specs)

        @functools.wraps(func)
        def check(*args, **kwargs):
            _df: AnnotatedDataFrame = _get_dataframe(*args, **kwargs)
            # Ensure that name mapping are applied correctly
            assert isinstance(args[0], PipelineElement)

            pipeline_element = args[0]
            fulfillment = _df.get_fulfillment(expected_specs=pipeline_element.input_specs)
            if fulfillment.is_met():
                if hasattr(pipeline_element, "parent_pipeline"):
                    getattr(pipeline_element, "parent_pipeline").on_transform_start(step=pipeline_element,
                                                                                    adf=_df)
                return func(*args, **kwargs)

            raise RuntimeError(
                "Input requirements are not fulfilled:" f" -- {fulfillment}"
            )

        return check

    return decorator


def output(requirements: Dict[str, Any]):
    """
    Specify the output for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param requirements: List of input requirements
    """
    required_output_specs = DataSpecification.from_requirements(
        requirements=requirements
    )

    def decorator(func):
        return_type = inspect.signature(func).return_annotation
        if return_type != AnnotatedDataFrame:
            raise RuntimeError(
                "output: decorator requires 'AnnotatedDataFrame' to be returned by function"
            )

        setattr(func, DECORATED_OUTPUT_SPECS, required_output_specs)

        @functools.wraps(func)
        def check(*args, **kwargs) -> AnnotatedDataFrame:
            setattr(func, DECORATED_OUTPUT_SPECS, required_output_specs)

            _df: AnnotatedDataFrame = _get_dataframe(*args, **kwargs)
            input_columns = list(_df.column_names)

            adf: AnnotatedDataFrame = func(*args, **kwargs)
            if adf is None:
                raise RuntimeError(
                    f"output: decorated function {func} must return 'AnnotatedDataFrame',"
                    f" but was 'None'"
                )

            if not isinstance(adf, AnnotatedDataFrame):
                raise RuntimeError(
                    f"output: decorated function {func} must return 'AnnotatedDataFrame', but was '"
                    f"{type(adf)}"
                )

            for c in input_columns:
                if c not in adf.column_names:
                    raise RuntimeError(
                        f"output: column '{c}' was removed by decorated function."
                        f" Only adding of columns is permitted."
                    )

            # Ensure that name mapping are applied correctly
            assert isinstance(args[0], PipelineElement)
            pipeline_element = args[0]

            parent_pipeline = None
            if hasattr(pipeline_element, "parent_pipeline"):
                parent_pipeline = getattr(pipeline_element, "parent_pipeline")

            try:
                # Ensure that metadata is up to date with the dataframe
                adf.update(expectations=pipeline_element.output_specs)
            except RuntimeError as e:
                txt = f"Failed to update metadata in pipeline element: {pipeline_element}"
                if parent_pipeline:
                    txt += f" in pipeline '{parent_pipeline.name}' ({type(parent_pipeline)})"

                raise RuntimeError(txt) from e

            if parent_pipeline:
                parent_pipeline.on_transform_end(step=pipeline_element,
                                                 adf=adf)

            return adf

        return check

    return decorator


def artifacts(requirements: Dict[str, Any]):
    """
    Specify the output for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param requirements: List of input requirements
    """
    required_artifact_specs = ArtifactSpecification(requirements=requirements)

    def decorator(func):
        return_type = inspect.signature(func).return_annotation
        if return_type != AnnotatedDataFrame:
            raise RuntimeError(
                "artifacts: decorator requires 'AnnotatedDataFrame' to be returned by function"
            )

        setattr(func, DECORATED_ARTIFACT_SPECS, required_artifact_specs)

        # When a pipeline does generate artifacts, then it might not provide any output, but serves only as
        # passthrough element. Hence, per default set an empty output spec if there is none.
        if not hasattr(func, DECORATED_OUTPUT_SPECS):
            setattr(
                func,
                DECORATED_OUTPUT_SPECS,
                DataSpecification.from_requirements(requirements={}),
            )

        @functools.wraps(func)
        def check(*args, **kwargs) -> AnnotatedDataFrame:
            result = func(*args, **kwargs)

            if not isinstance(args[0], PipelineElement):
                raise RuntimeError(
                    f"{args[0].__class__.__name__} must inherit from PipelineElement"
                )

            # Validate the spec with respect to the existing parent pipeline
            instance = args[0]
            try:
                required_artifact_specs.validate(
                    base_dir=instance.parent_pipeline.base_dir
                )
            except RuntimeError as e:
                raise RuntimeError(
                    f"artifacts: {func} is expected to generate an artifact. "
                    f" Pipeline element ran as part of pipeline: '{instance.parent_pipeline.name}'"
                ) from e

            return result

        return check

    return decorator


class PipelineElement(Transformer):
    """
    Allow to get the reference to a parent pipeline
    """

    #: Pipeline in which context this processor will be run
    parent_pipeline: DataProcessingPipeline

    #: Map names of input and outputs for a particular pipeline
    _name_mappings: Dict[str, str]

    def set_parent(self, pipeline: DataProcessingPipeline):
        """
        Sets the parent pipeline for this pipeline element

        :param pipeline: Parent pipeline
        """
        self.parent_pipeline = pipeline

    @property
    def name_mappings(self) -> Dict[str, str]:
        """
        Get current name mappings for this instance
        """
        if not hasattr(self, "_name_mappings"):
            self._name_mappings = {}
        return self._name_mappings

    @property
    def parameters(self) -> Dict[str, str]:
        """
        Get the current list of parameter for initialization
        """
        if not hasattr(self, "_parameters"):
            self._parameters = None

        self.prepare_parameters()

        return self._parameters

    def prepare_parameters(self):
        parameters = {}
        for p in inspect.signature(self.__init__).parameters:
            parameter_name = p
            if not parameter_name in ['args', 'kwargs']:
                # only process named parameters
                try:
                    parameters[parameter_name] = getattr(self, parameter_name)
                except AttributeError as e:
                    raise ValueError(f"PipelineElement: please ensure that {self.__class__.__name__}.__init__ keyword arguments"
                            f" are saved in an attribute of the same name: missing '{parameter_name}'")

        self._parameters = parameters


    def get_name(self, name: str) -> Any:
        """
        Add the fully resolved input/output name for this key.

        :param name: Name as used in the input spec, or pattern "{{x}}_suffix" in order to create a dynamic
                     output based an existing and renameable input
        :return: Name for this input after resolving name mappings and references
        """
        if not isinstance(name, str):
            raise TypeError(f"{self.__class__.__name__}.get_name: provided transformer label is not a string: {name}")

        if name in self.name_mappings:
            # allow multiple levels of name resolution, e.g.,
            # x -> y, y -> z --> x -> z
            mapped_name = self.name_mappings[name]
            if mapped_name == name:
                return name

            return self.get_name(mapped_name)

        # Allow to use patterns, so that an existing input
        # reference can be reused for dynamic labelling
        while re.search("{{\\w+}}", name):
            for match in re.finditer("{{\\w+}}", name):
                resolved_name = match.group()[2:-2]
                if resolved_name in self.name_mappings:
                    resolved_name = self.name_mappings[resolved_name]

                name = name.replace(match.group(), resolved_name)
        return name

    @abstractmethod
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Default transform implementation
        """

    @property
    def input_specs(self) -> List[DataSpecification]:
        if not hasattr(self.transform, DECORATED_INPUT_SPECS):
            raise AttributeError(
                f"{self.__class__.__name__}.validate: missing input specification"
                f" for processing step '{self.__class__.__name__}'"
            )

        generic_spec = getattr(self.transform, DECORATED_INPUT_SPECS)
        specs = copy.deepcopy(generic_spec)
        for spec in specs:
            spec.name = self.get_name(spec.name)

        return specs

    @property
    def output_specs(self) -> List[DataSpecification]:
        if not hasattr(self.transform, DECORATED_OUTPUT_SPECS):
            raise AttributeError(
                f"{self.__class__.__name__}.validate: missing output specification"
                f" for processing step '{self.__class__.__name__}'"
            )

        generic_spec = getattr(self.transform, DECORATED_OUTPUT_SPECS)
        specs = copy.deepcopy(generic_spec)
        for spec in specs:
            spec.name = self.get_name(spec.name)

        return specs

    @classmethod
    def create_new(cls,
                   module_name: str,
                   class_name: str,
                   name_mappings: Optional[Dict[str, Any]] = None,
                   parameters: Optional[Dict[str, Any]] = {}) -> PipelineElement:
        """
        Create a new PipelineElement Subclass instance dynamically

        :param module_name: Name of the module for the PipelineElement class
        :param class_name: Name of the PipelineElement subclass
        :param name_mappings: Dictionary of name mappings that should apply
        :return: Instance for the PipelineElement instance

        :raise ValueError: If module or class with given name is not specified
        :raise ImportError: If class could not be loaded
        """
        if module_name is None:
            raise ValueError(f"{cls.__name__}.create_new: missing 'module_name'")

        if class_name is None:
            raise ValueError(f"{cls.__name__}.create_new: missing 'class_name'")

        p_module = importlib.import_module(module_name)
        if hasattr(p_module, class_name):
            klass = getattr(p_module, class_name)
        else:
            raise ImportError(f"{cls.__name__}.create_new: could not load '{class_name}' from '{p_module}'")
        if parameters:
            instance = klass(**parameters)
        else:
            instance = klass()

        if name_mappings is None:
            name_mappings = {}
        instance._name_mappings = name_mappings
        return instance

    def __iter__(self):
        yield "module_name", f"{self.__class__.__module__}"
        yield "class_name", f"{self.__class__.__qualname__}"
        yield "parameters", self.parameters
        yield "name_mappings", self.name_mappings

    def __eq__(self, other):
        return dict(self) == dict(other)

    @classmethod
    def get_types(cls) -> List[PipelineElement]:
        """
        Get all available PipelineElement implementations

        :return: List of PipelineElement classes
        """
        klasses = []
        for subclass in cls.__subclasses__():
            klasses.append(subclass)
            klasses.extend(subclass._subclasses())
        return klasses

    @classmethod
    def _subclasses(cls) -> List[PipelineElement]:
        """
        Generate the list of subclasses for the calling class
        """
        klasses = []
        for subclass in cls.__subclasses__():
            klasses.append(subclass)
            klasses.extend(subclass._subclasses())
        return klasses

    @classmethod
    def generate_subclass_documentation(cls) -> str:
        """
        Generate the documentation for all subclasses of ::class::`PipelineElement`
        """
        implementations = sorted(cls.get_types(), key=str)
        txt = ""
        for k in implementations:
            txt += "=" * 80
            txt += f"\n{k.__name__} -- from {k.__module__}\n"
            txt += k.__doc__
            txt += '\n'
        return txt


class DataProcessingPipeline(PipelineElement):
    """
    A data-processing pipeline for a sequence of transformers

    :param name: Name of the pipeline
    :param base_dir: Base directory towards which transformer output which be relative
    :param steps: Steps that form the basis for this pipeline
    :param inplace_transformation: If true, the input :class:`damast.core.dataframe.AnnotatedDataFrame` is not
        copied when calling :func:`transform`. Else input data-frame is untouched
    :param name_mappings: Name mappings that should be applied to individual transformations.

    :raises ValueError: If any of the transformer names are `None`
    :raises AttributeError: If the transformer is missing the :func:`transform` function
    :raises AttributeError: If transformer is missing input or output decoration
    :raises RuntimeError: If the sequence of transformers does not satisfy the sequential requirements
    """

    #: Name of the processing pipeline
    name: str
    description: str

    #: Base path (which is forwarded to transformers, when calling
    #: transform)
    base_dir: Path

    #: The output specs - as specified by decorators
    _output_specs: List[DataSpecification]

    #: Check if the pipeline is ready to be run
    is_ready: bool
    # The processing steps that define this pipeline
    steps: List[Tuple[str, PipelineElement]]

    _inplace_transformation: bool
    _name_mappings: Dict[str, str]
    _processing_stats: Dict[str, Dict[str, Any]]

    _meta: Dict[str, str]

    def __init__(self, *,
                 name: str,
                 description: str = "",
                 base_dir: Union[str, Path] = tempfile.gettempdir(),
                 steps: List[Tuple[str, Union[Dict[str, Any], PipelineElement]]] = [],
                 inplace_transformation: bool = False,
                 name_mappings: Dict[str, str] = {},
                 meta: Dict[str, str] | None = None,
                 ):
        super().__init__()

        self.name = name
        self.description = description
        self.base_dir = Path(base_dir)

        self._output_specs = []
        self._inplace_transformation = inplace_transformation

        self._name_mappings = name_mappings
        self._processing_stats = {}

        if steps is None:
            raise ValueError(f"{self.__class__.__name__}.__init__:"
                             " steps must not be None")

        self.steps = []
        for step in steps:
            name, instance = step
            if isinstance(instance, PipelineElement):
                self.steps.append(step)
            elif isinstance(instance, dict):
                self.steps.append([step[0], PipelineElement.create_new(**step[1])])
            else:
                raise ValueError(f"{self.__class__.__name__}.__init__: could not instantiate PipelineElement"
                                 f" from {type(step)}")
        self.is_ready = False

        if meta is not None:
            self._meta = meta
        else:
            self._meta = {
                'damast_version': damast.version.__version__
            }

    @property
    def output_specs(self):
        if not self.is_ready:
            raise RuntimeError(
                f"{self.__class__.__name__}.output_specs: pipeline is not yet ready to run. "
                f"Please call 'prepare()' to set the correct output specs"
            )

        return self._output_specs

    def add(
            self,
            name: str,
            transformer: PipelineElement,
            name_mappings: Optional[Dict[str, str]] = None,
    ) -> DataProcessingPipeline:
        """
        Add a pipeline step

        :param name: Name of the step
        :param transformer: The transformer that shall be executed
        :param name_mappings: Allow to define a name mapping for this pipeline element instance
        :return:
        """
        transformer.set_parent(pipeline=self)
        if name_mappings is not None:
            transformer._name_mappings = name_mappings

        self.steps.append([name, transformer])
        self.is_ready = False
        return self

    @classmethod
    def validate(
            cls, steps: List[Tuple[str, PipelineElement]], metadata: MetaData
    ) -> Dict[str, Any]:
        """
        Validate the existing pipeline and collect the minimal input and output data specification.

        :param steps: processing steps
        :param metadata: the input metadata for this pipeline
        :return: The minimal output specification for this pipeline
        """
        # Keep track of the expected (minimal) specs at each step in the pipeline
        current_specs: Optional[List[DataSpecification]] = copy.deepcopy(
            metadata.columns
        )
        for name, transformer in steps:
            if name is None:
                raise ValueError(
                    f"{cls.__name__}.validate: missing name processing step"
                )

            if not hasattr(transformer, "transform"):
                raise AttributeError(
                    f"{cls.__name__}.validate: processing step '{name}' does not fulfill the"
                    f" TransformerMixin requirements - no method 'fit_transform' found"
                )

            input_specs = transformer.input_specs
            output_specs = transformer.output_specs

            md = MetaData(columns=current_specs, annotations=[])
            fulfillment = md.get_fulfillment(expected_specs=input_specs)
            if not fulfillment.is_met():
                raise RuntimeError(
                    f"{cls.__name__}.validate: Input requirements are not fulfilled. "
                    f"Current input available (at step '{name}'): {fulfillment}"
                )

            current_specs = DataSpecification.merge_lists(current_specs, input_specs)
            current_specs = DataSpecification.merge_lists(current_specs, output_specs)

        return {"steps": steps, "output_spec": current_specs}

    def to_str(self, indent_level: int = 0) -> str:
        """
        Output pipeline as string.

        :param indent_level: Indentation per step.
            It is multiplied by :attr:`damast.core.formatting.DEFAULT_INDENT`.
        :returns: The pipeline in string representation
        """
        hspace = DEFAULT_INDENT * indent_level

        data = hspace + self.__class__.__name__ + "\n"
        for step in self.steps:
            name, transformer = step
            data += hspace + DEFAULT_INDENT + name + ":\n"

            if hasattr(transformer.transform, DECORATED_DESCRIPTION):
                description = getattr(transformer.transform, DECORATED_DESCRIPTION)
                data += (
                        hspace + DEFAULT_INDENT * 2 + "description: " + description + "\n"
                )

            data += hspace + DEFAULT_INDENT * 2 + "input:\n"
            data += DataSpecification.to_str(
                transformer.input_specs, indent_level=indent_level + 4
            )

            data += hspace + DEFAULT_INDENT * 2 + "output:\n"
            data += DataSpecification.to_str(
                transformer.output_specs, indent_level=indent_level + 4
            )

        return data

    def save(self, dir: Union[str, Path]) -> Path:
        """
        Save the processing pipeline

        :param dir: directory where to save this pipeline
        """
        base_dir = Path(dir)
        base_dir.mkdir(parents=True, exist_ok=True)
        filename = base_dir / f"{self.name}{DAMAST_PIPELINE_SUFFIX}"

        with open(filename, "w") as f:
            yaml.dump(dict(self), f)
        return filename

    def __iter__(self):
        yield "name", self.name
        yield "description", self.description
        yield "meta", self._meta
        yield "base_dir", str(self.base_dir)
        yield "steps", [[step_name, dict(pipeline_element)] for step_name, pipeline_element in self.steps]

    def save_state(self,
                   df: AnnotatedDataFrame,
                   dir: Union[str, Path]) -> Path:
        """
        Save the processing pipeline

        :param dir: directory where to save this pipeline
        """
        filename = Path(dir) / f"{self.name}{VAEX_STATE_SUFFIX}"
        #df._dataframe.state_write(file=filename)
        df._dataframe.serialize(filename)
        return filename

    @classmethod
    def load(cls, path: Union[str, Path], name: str = "*") -> DataProcessingPipeline:
        """
        Load a :class:`DataProcessingPipeline` from file (without suffix :attr:`DAMAST_PIPELINE_SUFFIX`)

        :param path: Directory containing pipeline(s) (with fiven suffix) or pipeline filename
        :name: Name of file(s) without suffix. Can be a Regex expression
        """
        basename = f"{name}{DAMAST_PIPELINE_SUFFIX}"
        path = Path(path)
        if path.is_dir():
            files = list(path.glob(basename))
        elif path.is_file():
            files = [path]
        else:
            raise RuntimeError("{self.__class__.__name__}.load: {path}"
                               " is neither directory nor file")

        if len(files) == 1:
            filename = files[0]
        elif len(files) == 0:
            raise FileNotFoundError(
                f"{cls.__name__}.load: could not find '{basename}' in {dir}"
            )
        else:
            raise RuntimeError(
                f"{cls.__name__}.load: multiple pipeline available, pls be more specific:"
                f" {','.join([x.name for x in files])}"
            )

        with open(filename, "r") as f:
            data = yaml.load(f, Loader=yaml.SafeLoader)

        return cls(**data)

    @classmethod
    def load_state(
            cls, df: AnnotatedDataFrame, dir: Union[str, Path], name: str = "*"
    ) -> AnnotatedDataFrame:
        """
        Load a ``vaex`` state (from file) to a :class:`damast.core.AnnotatedDataFrame`.

        :param df: The data-frame
        :param dir: Directory of vaex state file
        :param name: Regex describing vaex state file (without :attr:`VAEX_STATE_SUFFIX`)
        :raises FileNotFoundError: If no files found
        :raises RuntimeError: If multiple vaex-states satisfies the ``name`` requirement.
        :return: The data-frame with applied state
        """
        basename = f"{name}{VAEX_STATE_SUFFIX}"
        dir_path = Path(dir)

        files = list(dir_path.glob(basename))
        if len(files) == 1:
            filename = files[0]
        elif len(files) == 0:
            raise FileNotFoundError(
                f"{cls.__name__}.load: could not find '{basename}' in {dir_path}"
            )
        else:
            raise RuntimeError(
                f"{cls.__name__}.load: multiple pipeline available, pls be more specific:"
                f" {','.join([x.name for x in files])}"
            )

        #df.dataframe.state_load(file=filename)
        df._dataframe = df.dataframe.deserialize(filename)
        return df

    def prepare(self, df: AnnotatedDataFrame) -> DataProcessingPipeline:
        """
        Prepare the pipeline by applying necessary name mapping and validating all steps that define the pipeline.

        A :class:`DataProcessingPipeline` must be prepared before execution.

        :returns: This instance of data processing pipeline
        """

        # At this stage, ensure that the dataframe conforms to the metadata
        try:
            df.validate_metadata()
        except ValueError as e:
            raise RuntimeError(
                f"{self.__class__.__name__}.prepare: specification of the provided AnnotatedDataFrame"
                f" does not match its data. If you modified the frame after construction ensure"
                f" consistency between 'data' and 'metadata' by using "
                f" AnnotatedDataFrame.validate_metadata()"
                f" -- {e}"
            ) from e

        # The pipeline will define a name mapping (only for items in its interface)
        for k, v in self._name_mappings.items():
            for step in self.steps:
                _, pipeline_element = step

                input_columns = [x.name for x in pipeline_element.input_specs]
                output_columns = [x.name for x in pipeline_element.output_specs]

                columns = input_columns + output_columns
                for x in columns:
                    if x == k:
                        pipeline_element.name_mappings[x] = v

        validation_result = self.validate(steps=self.steps, metadata=df.metadata)
        self.is_ready = True

        self.steps: List[Tuple[str, PipelineElement]] = validation_result["steps"]
        self._output_specs: List[DataSpecification] = validation_result["output_spec"]

        return self

    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Apply pipeline on given annotated dataframe

        .. note::
            If any filters are applied by the pipeline, these are executed at the end of the
            transformation.

        :param df: The input dataframe
        :returns: The transformed dataframe
        """
        if not isinstance(df, AnnotatedDataFrame):
            raise TypeError(
                f"{self.__class__.__name__}.transform"
                f" expected an annotated dataframe for processing, "
                f"got {type(df)}"
            )

        if not self.is_ready:
            self.prepare(df=df)

        if self._inplace_transformation:
            in_df = df
        else:
            in_df = copy.deepcopy(df)

        steps = [t for _, t in self.steps]
        pipeline = Pipeline(steps)

        if df.is_empty():
            raise RuntimeError(
                f"{self.__class__.__name__}.transform: there is no data available to transform"
            )

        adf = pipeline.transform(df=in_df)
        assert isinstance(adf, AnnotatedDataFrame)

        adf.validate_metadata(validation_mode=damast.core.ValidationMode.UPDATE_METADATA)
        return adf

    def on_transform_start(self,
                           step: PipelineElement,
                           adf: AnnotatedDataFrame):
        """
        Default implementation of the on_transform_start callback.

        This output is a message in INFO log level
        """
        if hasattr(step, "transform_start"):
            return

        _log.info(f"[transform] start: {step.__class__.__name__} - {step.name_mappings}")
        start_time = datetime.now(timezone.utc)
        setattr(step, "transform_start", start_time)

        step_name = [x for x, y in self.steps if y == step][0]
        self._processing_stats[step_name] = {
            "input_dataframe_length": adf.shape[0],
            "start_time": start_time
        }

    def on_transform_end(self,
                         step: PipelineElement,
                         adf: AnnotatedDataFrame):
        """
        Default implementation of the on_transform_end callback.

        This outputs a message in INFO log level
        """
        if not hasattr(step, "transform_start"):
            return

        start = getattr(step, "transform_start")
        end_time = datetime.now(timezone.utc)
        delta = (end_time - start).total_seconds()
        delattr(step, "transform_start")
        _log.info(f"[transform] end: {step.__class__.__name__} - {step.name_mappings}: "
                  f"{delta} seconds, {adf.shape[0]} remaining rows)")

        step_name = [x for x, y in self.steps if y == step][0]
        self._processing_stats[step_name] = {
            "processing_time_in_s": delta,
            "output_dataframe_length": adf.shape[0],
            "end_time": end_time
        }

    def __repr__(self) -> str:
        """
        Create the representation string for this :class:`DataProcessingPipeline`.

        :returns: String representation
        """
        return self.to_str()

def polar_data_type_constructor(loader, tag_suffix, node):
    module_whitelist = ["polars.datatypes.classes"]
    for m in module_whitelist:
        if tag_suffix.startswith(m):
            module = importlib.import_module(m)
            object_name = tag_suffix.replace(m + ".","")
            return getattr(module, object_name)

    raise ValueError(f"damast.core.dataprocessing: no constructor registered for {tag_suffix=}")

yaml.SafeLoader.add_multi_constructor("tag:yaml.org,2002:python/name:", polar_data_type_constructor)



"""
Module that contains transformers which do not modify data, but serve only data exploration.
"""
from pathlib import Path
from typing import Union

from damast.core.units import units
from damast.data_handling.exploration import PLOT_DPI, plot_histograms, plot_lat_lon

__all__ = [
    "BaseVisualiser",
    "PlotHistograms",
    "PlotLatLon",
]

import damast.core
from damast.core.dataprocessing import AnnotatedDataFrame, PipelineElement


class BaseVisualiser(PipelineElement):

    #: The output directory for generated artifacts
    output_dir: Path
    filename_prefix: str
    dpi: int

    def __init__(self,
                 output_dir: Union[str, Path],
                 filename_prefix: str = '',
                 dpi: int = PLOT_DPI):
        self.output_dir = Path(output_dir)
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.filename_prefix = filename_prefix
        self.dpi = dpi


class PlotHistograms(BaseVisualiser):
    def __init__(self,
                 output_dir: Union[str, Path],
                 filename_prefix: str = "histogram"):
        super().__init__(output_dir=output_dir,
                         filename_prefix=filename_prefix)

    @damast.core.input({})
    @damast.core.output({})
    @damast.core.describe("Plot histograms of all columns in dataframe")
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        plot_histograms(df=df._dataframe,
                        output_dir=self.output_dir,
                        filename_prefix=self.filename_prefix)
        return df


class PlotLatLon(BaseVisualiser):
    def __init__(self,
                 output_dir: Union[str, Path],
                 filename_prefix: str = "lat-lon"):
        super().__init__(output_dir=output_dir,
                         filename_prefix=filename_prefix)

    @damast.core.input({"LAT": {"unit": units.deg}, "LON": {"unit": units.deg}})
    @damast.core.output({})
    @damast.core.describe("Plot Latitude and longitude ")
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:

        plot_lat_lon(df=df._dataframe,
                     output_dir=self.output_dir,
                     filename_prefix=self.filename_prefix)
        return df



"""
Module containing the BaseParser functionality, in order to simplify the usage of subparsers.
"""
import argparse
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path


@dataclass
class FilesStats:
    # Sum of file sizes in MB
    total_size: int
    files: list[str]

    @property
    def number_of_files(self):
        return len(self.files)


class BaseParser(ABC):

    @abstractmethod
    def __init__(self, parser: ArgumentParser):
        parser.add_argument('--active_subparser',
                            default=self,
                            action='store',
                            help=argparse.SUPPRESS)

    def execute(self, args):
        print(f"Subparser: {args.active_subparser.__class__.__name__}")
        pass

    def get_files_stats(self, files: list[str | Path]) -> FilesStats:
        """
        Get basic stats for the list of files given
        """
        expanded_pattern = []
        for file in files:
            base = Path(file).parent
            name = Path(file).name
            expanded_pattern += [x for x in base.glob(name)]

        sum_st_size = 0
        for idx, file in enumerate(expanded_pattern, start=1):
            sum_st_size += file.stat().st_size

        return FilesStats(total_size=round(sum_st_size / (1024**2), ndigits=0), files=expanded_pattern)
